# %%
import sys
from neurodsp.sim import sim_powerlaw, sim_oscillation
import numpy as np
import scipy.signal as dsp
import matplotlib.pyplot as plt
import seaborn as sns
import pandas as pd

sns.set_style('ticks')
sns.set_context('poster')
sys.path.append('/home/schmidtfa/git/SPRiNT')
from scipy.signal import ShortTimeFFT

from pyrasa.irasa import irasa_sprint
from pyrasa.utils.irasa_utils import _gen_time_from_sft

# %% now lets do time resolved irasa
fs = 500
n_seconds = 60
duration = 2
overlap = 0.5


sim_components = {'sim_powerlaw': {'exponent': -1}, 'sim_oscillation': {'freq': 10}}

# x1 = sim_combined(n_seconds=10, fs=fs, components=sim_components)
x1 = sim_oscillation(n_seconds=10, fs=fs, freq=10)  # * .5
x2 = sim_powerlaw(n_seconds=10, fs=fs, exponent=-1)
x3 = sim_oscillation(n_seconds=10, fs=fs, freq=20)
x4 = sim_oscillation(n_seconds=20, fs=fs, freq=30)


x = np.concatenate([x1 + x2, np.concatenate([x2, x2 + x1]) + x4, x2 + x3, x2])
x_2 = np.concatenate([np.concatenate([x2, x2 + x1]) + x4, x1 + x2, x2 + x3, x2])

x = np.tile(x, 10)
x_2 = np.tile(x_2, 10)

x_l = np.concatenate([x[np.newaxis, :], x_2[np.newaxis, :]], axis=0)

# %%
plt.plot(x[:10000])


# %%

win = dsp.windows.tukey(int(fs * duration), 0.25)
SFT = ShortTimeFFT(win, hop=int(win.shape[0]), fs=fs, scale_to='psd')

# get time and frequency info
my_time = _gen_time_from_sft(SFT, x_l)
freqs = SFT.f

sgramm = SFT.spectrogram(x_l, detr='constant')
tmin, tmax = SFT.extent(x_l.shape[-1])[:2]

# %%
plt.loglog(freqs, sgramm[0, :, 2])


# %%
freq_mask = freqs < 40
new_freqs = freqs[freq_mask]

import time

start = time.time()
sgramm_aperiodic, sgramm_periodic, freqs_ap, time_ap = irasa_sprint(
    x_l, fs=fs, band=(1, 100), hop=1, freq_res=0.5, win_duration=2, smooth=False, n_avgs=[3, 7, 11]
)
stop = time.time()
print(stop - start)
# %%
freq_mask_ap = freqs_ap < 40
new_freqs_ap = freqs_ap[freq_mask_ap]

# %%
plt.plot(freqs_ap, sgramm_periodic[0, :, 3])


# %%
for ch in [0, 1]:
    f, axes = plt.subplots(ncols=3, figsize=(12, 4))

    for ax in axes:
        ax.set_ylabel('Frequency (Hz)')
        ax.set_xlabel('Time (s)')

    axes[0].set_title('Combined \n spectrum')
    axes[0].imshow(
        sgramm[ch, freq_mask, :], extent=(tmin, tmax, new_freqs.min(), new_freqs.max()), origin='lower', aspect='auto'
    )

    axes[1].set_title('Periodic \n spectrum')
    axes[1].imshow(
        sgramm_periodic[ch, freq_mask_ap, :],
        extent=(tmin, tmax, new_freqs_ap.min(), new_freqs_ap.max()),
        origin='lower',
        aspect='auto',
    )

    axes[2].set_title('Aperiodic \n spectrum')
    axes[2].imshow(
        sgramm_aperiodic[ch, freq_mask_ap, :],
        extent=(tmin, tmax, new_freqs_ap.min(), new_freqs_ap.max()),
        origin='lower',
        aspect='auto',
    )

    plt.tight_layout()
# %%
from pyrasa.utils.peak_utils import get_peak_params_sprint, get_band_info

df_peaks = get_peak_params_sprint(sgramm_periodic, freqs=freqs_ap, times=time_ap, min_peak_height=0.2)
df_alpha_peaks = get_band_info(df_peaks, freq_range=(8, 12), ch_names=[0, 1]).query('ch_name == 0')
df_alpha_peaks.head(10)

# %%
from SPRiNT_py import SPRiNT_stft_py, SPRiNT_remove_outliers
# %%


opt = {
    'sfreq': fs,  # Input sampling rate
    'WinLength': duration,  # STFT window length
    'WinOverlap': 0,  # Overlap between sliding windows (in %)
    'WinAverage': 5,  # Number of overlapping windows being averaged
    'rmoutliers': 1,  # Apply peak post-processing
    'maxTime': 6,  # Maximum distance of nearby peaks in time (in n windows)
    'maxFreq': 2.5,  # Maximum distance of nearby peaks in frequency (in Hz)
    'minNear': 3,  # Minimum number of similar peaks nearby (using above bounds)
}

output = SPRiNT_stft_py(x_l, opt)

# %%
for ch in [0, 1]:
    f, axes = plt.subplots(ncols=2, figsize=(8, 4))

    for ax in axes:
        ax.set_ylabel('Frequency (Hz)')
        ax.set_xlabel('Time (s)')

    axes[0].set_title('Combined \n spectrum (SPRINT)')
    axes[0].imshow(
        output['TF'][ch, :, freq_mask],
        # extent=(tmin, tmax, new_freqs.min(), new_freqs.max()),
        origin='lower',
        aspect='auto',
    )

    axes[1].set_title('Combined \n spectrum (IRASA)')
    axes[1].imshow(
        output['TF'][ch, :, freq_mask],
        # extent=(tmin, tmax, new_freqs.min(), new_freqs.max()),
        origin='lower',
        aspect='auto',
    )

# %%
from fooof import FOOOFGroup, fit_fooof_3d
from fooof.objs.utils import combine_fooofs

start = time.time()
fg = FOOOFGroup(peak_width_limits=[1, 8], min_peak_height=0.2, max_n_peaks=3)
fgs = fit_fooof_3d(fg, output['freqs'], output['TF'])

stop = time.time()

# %%
print(stop - start)

df_fooof = pd.DataFrame(fgs[0].get_params('gaussian_params'), columns=('CF', 'PW', 'BW', 'Time'))
df_alpha_fooof = df_fooof.query('CF < 12').query('CF > 8')

# %%
plt.plot(df_alpha_peaks['time'], df_alpha_peaks['pw'])
# %%
plt.plot(df_alpha_fooof['Time'], df_alpha_fooof['PW'])
# %%
plt.plot(df_alpha_peaks['time'], df_alpha_peaks['cf'])
# %%
plt.plot(df_alpha_fooof['Time'], df_alpha_fooof['CF'])

# %%
plt.hist(df_alpha_fooof['CF'])
plt.xlim(9.5, 10.5)
# %%
plt.hist(df_alpha_peaks['cf'])
plt.xlim(9.5, 10.5)
# %%

# %%
fgs = [SPRiNT_remove_outliers(fgs[i], output['ts'], opt) for i in range(2)]
# %%
df_alpha_foof_cl = (
    pd.DataFrame(fgs[0].get_params('peak_params'), columns=('CF', 'PW', 'BW', 'Time')).query('CF < 12').query('CF > 8')
)

# %%
plt.plot(df_alpha_peaks['time'], df_alpha_peaks['pw'])
# %%
plt.plot(df_alpha_foof_cl['Time'], df_alpha_foof_cl['PW'])
# %%
plt.plot(df_alpha_peaks['time'], df_alpha_peaks['cf'])
# %%
plt.plot(df_alpha_foof_cl['Time'], df_alpha_foof_cl['CF'])

# %%
pd.DataFrame(fgs[0].get_params('peak_params'))

# %%
print(fgs[0].get_params('peak_params', 'CF'))

# %%
all_fg = combine_fooofs(fgs)

# Explore the results from across all model fits
all_fg.print_results()

# %%
all_fg.report()
# %%



#%%
import sys
from neurodsp.sim import sim_combined
import numpy as np
import seaborn as sns

sns.set_style('ticks')
sns.set_context('paper')
sys.path.append('../')


import matplotlib as mpl
new_rc_params = {'text.usetex': False,
                 "svg.fonttype": 'none'
}
mpl.rcParams.update(new_rc_params)

import mne
from mne.datasets import sample



from pyrasa.irasa_mne import irasa_raw, irasa_epochs

#%% now lets check the mne python implementation


data_path = sample.data_path()
subjects_dir = data_path / "subjects"
subject = "sample"

meg_path = data_path / "MEG" / "sample"
raw_fname = meg_path / "sample_audvis_raw.fif"
fwd_fname = meg_path / "sample_audvis-meg-eeg-oct-6-fwd.fif"
fwd = mne.read_forward_solution(fwd_fname)
src = fwd["src"]


raw = mne.io.read_raw_fif(raw_fname)
picks = mne.pick_types(raw.info, meg='mag', eeg=False, 
                       stim=False, eog=False, exclude="bads")
raw.pick(picks)

#%% lets simulate 2 different signals
info = raw.info.copy()
tstep = 1.0 / info["sfreq"]

alpha = sim_combined(60, 
                     info["sfreq"], 
                     components={'sim_oscillation': {'freq' : 10},
                                 'sim_powerlaw': {'exponent': -1.5,}})

beta = sim_combined(60, 
                     info["sfreq"], 
                     components={'sim_oscillation': {'freq' : 20},
                                 'sim_powerlaw': {'exponent': -1.5,}})


#%% now lets simulate some data
#TODO: put in a helper function

def simulate_raw_(signal, scaling_fator, region, subject, info, subjects_dir):
    
    '''Shorthand function to simulate a dipole'''

    selected_label = mne.read_labels_from_annot(
        subject, regexp=region, subjects_dir=subjects_dir,
    )[0]
    location = "center"  # Use the center of the region as a seed.
    extent = 10.0  # Extent in mm of the region.
    label = mne.label.select_sources(
        subject, selected_label, location=location, 
        extent=extent, subjects_dir=subjects_dir
    )

    #this is the source ts
    source_time_series = signal * scaling_fator

    n_events = 1
    events = np.zeros((n_events, 3), int)
    events[:, 0] = np.arange(n_events)  # Events sample.
    events[:, 2] = 1  # All events have the sample id.

    source_simulator = mne.simulation.SourceSimulator(src, tstep=tstep)
    source_simulator.add_data(label, source_time_series, events)

    raw = mne.simulation.simulate_raw(info, source_simulator, forward=fwd)
    #cov = mne.make_ad_hoc_cov(raw.info)
    #mne.simulation.add_noise(raw, cov, iir_filter=[0.2, -0.2, 0.04])

    return raw

#%% simulate alpha + beta and combine
raw_alpha = simulate_raw_(alpha, 5e-9, "caudalmiddlefrontal-rh", 
                          subject, info, subjects_dir)

raw_beta = simulate_raw_(beta, 10e-10, "caudalmiddlefrontal-lh", 
                          subject, info, subjects_dir)

raw_data = raw_alpha.get_data() + raw_beta.get_data()

raw = mne.io.RawArray(raw_data, info)

picks = mne.pick_types(raw.info, meg='mag', eeg=False, 
                       stim=False, eog=False, exclude="bads")
raw.pick(picks)

#%%
aperiodic_mne, periodic_mne = irasa_raw(raw, band=(.25, 50), 
                                duration=2, 
                                hset_info=(1.,2.,.05),
                                as_array=False)
#%%
df_slopes, df_gof = aperiodic_mne.get_slopes()

#%%
df_slopes.head()
#%%
df_gof.head()

#%%

peak_df = periodic_mne.get_peaks()

peak_df.head()

periodic_mne.plot()

#%% now lets check-out the events
event_id = {
    "Auditory/Left": 1,
    "Auditory/Right": 2,
    "Visual/Left": 3,
    "Visual/Right": 4,
}
tmin = -0.2
tmax = 0.5

# Load real data as the template
event_fname = meg_path / "sample_audvis_filt-0-40_raw-eve.fif"
events = mne.read_events(event_fname)


epochs = mne.Epochs(
    raw,
    events,
    event_id,
    tmin,
    tmax,
    #picks=picks,
    baseline=None,
    preload=True,
    verbose=False,
)

#%%

aperiodic, periodic = irasa_epochs(epochs, 
                                   band=(.5, 50), 
                                   hset_info=(1.,2.,.05), 
                                   as_array=False)

#%%
periodic.get_peaks(polyorder=0)

#%%
aperiodic.get_slopes()

#%%
periodic["Auditory/Left"].average().plot_topomap(dB=False)
#%%
periodic["Auditory/Right"].average().plot_topomap(dB=False)
#%%
periodic["Visual/Left"].average().plot_topomap(dB=False)
#%%
periodic["Visual/Right"].average().plot_topomap(dB=False)



#%%
aperiodic["Auditory/Left"].average().plot_topomap(dB=False)
#%%
aperiodic["Auditory/Right"].average().plot_topomap(dB=False)
#%%
aperiodic["Visual/Left"].average().plot_topomap(dB=False)
#%%
aperiodic["Visual/Right"].average().plot_topomap(dB=False)

# %%



#%%
import sys
from neurodsp.sim import sim_knee
from neurodsp.utils import create_times
import numpy as np
import scipy.signal as dsp
import matplotlib.pyplot as plt
import seaborn as sns
import pandas as pd

sns.set_style('ticks')
sns.set_context('talk')
sys.path.append('../../')

from irasa import irasa


#%%
# Simulation settings
# Set some general settings, to be used across all simulations
fs = 500
n_seconds = 60
duration=4
overlap=0.5

# Create a times vector for the simulations
times = create_times(n_seconds, fs)

#%%
exponent_1 = 0.
exponents = [0.5, 1, 2, 3]
knee_freq = 15


knee_ap = []
knee_osc = []
for exponent_2 in exponents:

    knee_ap.append(sim_knee(n_seconds, fs, 
                        exponent1=-1*exponent_1, 
                        exponent2=-1*exponent_2, 
                        knee=(knee_freq ** (2*exponent_1 + exponent_2))))
    
    # knee_osc.append(sim_combined(n_seconds, 
    #                             fs, 
    #                             components={'sim_oscillation': {'freq' : 10},
    #                                         'sim_knee': {'exponent1': -0., 
    #                                                     'exponent2':-1*exponent, 
    #                                                     'knee': knee_freq ** exponent}}))

    
# %% plot timeseries

tmax = times < 60
fig, ax = plt.subplots(figsize=(6,3))

ax.plot(times[tmax], knee_ap[0][tmax], label=f'slope = -{exponents[0]}')
ax.plot(times[tmax], knee_ap[1][tmax]+3, label=f'slope = -{exponents[1]}')
ax.plot(times[tmax], knee_ap[2][tmax]+6, label=f'slope = -{exponents[2]}')

sns.despine()
plt.legend()

# %% test irasa
kwargs_psd = {'window': 'hann',
              'nperseg': int(fs*duration), 
              'noverlap': int(fs*duration*overlap)}

# Calculate original spectrum
freq_ap, psds_ap = dsp.welch(knee_ap, fs=fs, **kwargs_psd)
fmax= 150
freq_rasa_ap, psd_aperiodics_ap, psd_periodics_ap = irasa(np.array(knee_ap), band=(0.1, fmax), 
                                                fs=fs, kwargs_psd=kwargs_psd,
                                                hset_info=(1.,2.,.01))


freq_mask = freq_ap < fmax

# fig, axes = plt.subplots(ncols=3, figsize=(12, 4))

# for ix, ax in enumerate(axes):
#     ax.loglog(freq_ap[freq_mask], psds_ap[ix, freq_mask])
#     ax.loglog(freq_rasa_ap, psd_aperiodics_ap[ix])


# %% now lets parametrize the fractal part
from aperiodic_utils import compute_slope, knee_model

#%%
cols = ['Knee', 'Offset', 'Exponent_1', 'Exponent_2']

aps_cmb, gof_cmb = compute_slope(psd_aperiodics_ap, freq_rasa_ap, fit_func='knee')

#%%
aps_cmb

#%%
gof_cmb

#%%
plt.loglog(freq_ap, psds_ap[0], 'ko')
plt.loglog(freq_ap, psds_ap[1], 'ko')
plt.loglog(freq_ap, psds_ap[2], 'ko')
plt.loglog(freq_rasa_ap, 10**knee_model(freq_rasa_ap, *aps_cmb.iloc[0].to_numpy()[:-3]))
plt.loglog(freq_rasa_ap, 10**knee_model(freq_rasa_ap, *aps_cmb.iloc[1].to_numpy()[:-3]))
plt.loglog(freq_rasa_ap, 10**knee_model(freq_rasa_ap, *aps_cmb.iloc[2].to_numpy()[:-3]))
plt.axvline(knee_freq)
#%%

aps1, gof1 = compute_slope(psds_ap[0][np.newaxis, 1:], freq_ap[1:], fit_func='knee')
aps2, gof2 = compute_slope(psds_ap[1][np.newaxis, 1:], freq_ap[1:], fit_func='knee')
aps3, gof3 = compute_slope(psds_ap[2][np.newaxis, 1:], freq_ap[1:], fit_func='knee')

aps_cmb = pd.concat([aps1, aps2, aps3])

aps_cmb
#%%
#%%
plt.loglog(freq_ap, psds_ap[0], 'ko')
plt.loglog(freq_ap, psds_ap[1], 'ko')
plt.loglog(freq_ap, psds_ap[2], 'ko')
plt.loglog(freq_ap, 10**knee_model(freq_ap, *aps1.to_numpy()[0][:-3]))
plt.loglog(freq_ap, 10**knee_model(freq_ap, *aps2.to_numpy()[0][:-3]))
plt.loglog(freq_ap, 10**knee_model(freq_ap, *aps3.to_numpy()[0][:-3]))
plt.axvline(knee_freq)

 
#%%
cur_id = 0
cumsum_psd = np.cumsum(psds_ap[cur_id][1:])
half_pw_freq = freq_ap[1:][np.abs(cumsum_psd - (0.5 * cumsum_psd[-1])).argmin()]
#exp_guess = np.abs(psds_ap[cur_id][1] / psds_ap[cur_id][-1]) / (np.log10(freq_ap[-1] / freq_ap[1]))
exp_guess = np.log10(psds_ap[cur_id][1] / psds_ap[cur_id][-1])# / 2 #/ np.log10(freq_ap[-1] - freq_ap[1]))
exp_guess

#%%
plt.plot(np.log10(freq_ap[1:]), np.log10(psds_ap[cur_id][1:].min()) * -1 + np.log10(psds_ap[cur_id][1:]))
plt.axvline(np.log10(half_pw_freq))
knee_freq ** (exp_guess)

#%%
from scipy.optimize import curve_fit

def expo_reg(x, a, k, b):

    '''Uses specparams fitting function with a knee'''

    y_hat = a - np.log10(k + x**b)

    return y_hat


def expo_reg_nk(x, a, b):

    '''Uses specparams fitting function without a knee'''

    y_hat = a - np.log10(x**b)

    return y_hat


def knee_model(x, a, k, b1, b2):

    y_hat = a - np.log10(x**b1 * (k + x**b2))

    return y_hat


def mixed_model(x, x1, a1, b1, k, a2, b2, b3):

    '''
    Fit the data using a piecewise function. 
    Where Part A is a fixed model fit and Part B is a fit that allows for a knee.
    Use this to model aperiodic activity that contains a spectral plateau.
    NOTE: This might require some additional testing action
    '''

    condlist = [x < x1,  x > x1]
    funclist = [lambda x: a1 - np.log10(x**b1), 
                lambda x: a1 - np.log10((x-x1)**b2 * (k + (x-x1)**b3))]
    return np.piecewise(x, condlist, funclist)


#%%
cur_id = 2

curv_fit_kwargs = {'maxfev': 5000,
                   'ftol': 1e-5, 
                   'xtol': 1e-5, 
                   'gtol': 1e-5,}
off_guess = [psds_ap[cur_id][1]]
exp_guess = [np.log10(psds_ap[cur_id][1] / psds_ap[cur_id][-1]) - np.log10(freq_ap[-1] / freq_ap[1])]
cumsum_psd = np.cumsum(psds_ap[cur_id][1:])
half_pw_ix = np.abs(cumsum_psd - (0.5 * cumsum_psd[-1])).argmin()
half_pw_freq = freq_ap[half_pw_ix] 
half_pw_off = [psds_ap[cur_id][half_pw_ix]]
#make the knee guess the point where we have half the power in the spectrum seems plausible to me
knee_guess = [half_pw_freq ** exp_guess[0]] #convert knee freq to knee val
curv_fit_kwargs['p0'] = np.array(knee_guess + off_guess + exp_guess + half_pw_off + knee_guess + exp_guess + exp_guess) #TODO: Really pay attention here
curv_fit_kwargs['bounds'] = ((-np.inf, -np.inf, -np.inf, -np.inf, -np.inf, -np.inf, -np.inf), 
                             (np.inf, np.inf, np.inf, np.inf, np.inf, np.inf, np.inf)) 

p, _ = curve_fit(knee_model, freq_ap[1:], np.log10(psds_ap[cur_id][1:]))
p2, _ = curve_fit(mixed_model, freq_ap[1:], np.log10(psds_ap[cur_id][1:]), **curv_fit_kwargs)

plt.loglog(freq_ap[1:], psds_ap[cur_id][1:], 'ko')
plt.loglog(freq_ap[1:], 10**knee_model(freq_ap[1:], *p))
plt.loglog(freq_ap[1:], 10**mixed_model(freq_ap[1:], *p2))


#%% now do the same with oscillations
from peak_utils import get_peak_params

freq, psds = dsp.welch(knee_osc, fs=fs, **kwargs_psd)
freq_rasa, psd_aperiodics, psd_periodics = irasa(np.array(knee_osc), 
                                                 band=(0.1, 100),
                                                 fs=fs,
                                                 kwargs_psd=kwargs_psd,
                                                 hset_info=(1.,2.,.01))

freq_mask = freq < 100

fig, axes = plt.subplots(ncols=3, figsize=(12, 4))

for ix, ax in enumerate(axes):
    ax.loglog(freq[freq_mask], psds[ix, freq_mask])
    ax.loglog(freq_rasa, psd_aperiodics[ix])

get_peak_params(psd_periodics, 
                freq_rasa,
                min_peak_height=0.01,
                peak_width_limits=(0.25, 8),
                peak_threshold=1)

# %%
plt.plot(psd_periodics[0,:])
plt.plot(psd_periodics[1,:])
plt.plot(psd_periodics[2,:])





# %%
import joblib
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from pathlib import Path
import numpy as np

import scipy.stats as stats

sns.set_style('ticks')
sns.set_context('poster')

# %%
path_list = list(Path('/home/schmidtfa/git/pyrasa/paper/data/knee_sim').glob('*/*.dat'))
# %%

all_dfs = []
for f in path_list:
    f_info = str(f).split('/')[-1].split('_')
    cur_data = pd.concat(joblib.load(f)['ap_fits'])
    cur_data['pre_knee_exp'] = float(f_info[8])
    cur_data['n_seconds'] = int(f_info[-5])
    cur_data['fit_func'] = f_info[4]
    all_dfs.append(cur_data)

# %%
cur_pre = 0.5
n_secs = 300

df_cmb = pd.concat(all_dfs).query(f'pre_knee_exp == {cur_pre}')
exponent = 'Exponent_2'  # or Exponent_2
# %%

# df_cmb['Knee Frequency (Hz)'] = df_cmb['Knee'] ** (1. / (2*cur_pre + df_cmb[exponent]))
df_cmb['relative Error (Knee Frequency (Hz))'] = df_cmb['GT_Knee_Freq'] - df_cmb['Knee Frequency (Hz)']
df_cmb['relative Error (Exponent)'] = cur_pre + df_cmb['GT_Exponent'] - df_cmb[exponent]
df_cmb['Error (Knee Frequency (Hz))'] = np.abs(df_cmb['GT_Knee_Freq'] - df_cmb['Knee Frequency (Hz)'])
df_cmb['Error (Exponent)'] = np.abs(cur_pre + df_cmb['GT_Exponent'] - df_cmb[exponent])
df_cmb = df_cmb[df_cmb['Knee Frequency (Hz)'] < 100]

# %%
sns.catplot(df_cmb, x='n_seconds', y='Error (Knee Frequency (Hz))', hue='param_type', kind='point')
# %%
sns.catplot(df_cmb, x='n_seconds', y='Error (Exponent)', hue='param_type', kind='point')


df_cmb_cut = df_cmb.query(f'n_seconds == {n_secs}').drop(columns='n_seconds')

no_rasa = df_cmb_cut.query('param_type == "norasa"')
irasa = df_cmb_cut.query('param_type == "irasa"')
# %%
cor_cols = [
    'Offset',
    'Exponent_1',
    'Exponent_2',
    'Knee Frequency (Hz)',
    'relative Error (Knee Frequency (Hz))',
    'relative Error (Exponent)',
]

for col in cor_cols:
    print(f'{col}: {stats.pearsonr(irasa[col], no_rasa[col])}')

# %%
df_list = []
for sec in [4, 8, 30, 60, 180, 300]:
    df_cmb_cut = df_cmb.query(f'n_seconds == {sec}').drop(columns='n_seconds')

    no_rasa = df_cmb_cut.query('param_type == "norasa"')
    irasa = df_cmb_cut.query('param_type == "irasa"')

    for col in cor_cols:
        cur_df = pd.DataFrame({'r': stats.pearsonr(irasa[col], no_rasa[col])[0], 'col': col, 'nsecs': sec}, index=[0])
        df_list.append(cur_df)

# %%
df_corrs = pd.concat(df_list)

grid = sns.FacetGrid(df_corrs, col='col', hue='col', palette='deep', col_wrap=3, height=4, aspect=2)

grid.map(plt.plot, 'nsecs', 'r', marker='o')
grid.set(ylim=(0, 1.1))
grid.tight_layout(w_pad=1)

# %%
cols_of_interest = [
    'param_type',
    'Knee',
    'Offset',
    'Exponent_2',
    'Error (Knee Frequency (Hz))',
    'Error (Exponent)',
    'Knee Frequency (Hz)',
]


corr_eff = df_cmb_cut[cols_of_interest].groupby('param_type').corr()

# %%
f, axes = plt.subplots(ncols=2, figsize=(20, 10), sharex=True, sharey=True)

sns.heatmap(corr_eff.query('param_type == "irasa"'), ax=axes[0])
sns.heatmap(corr_eff.query('param_type == "norasa"'), ax=axes[1])

for ax in axes:
    ax.set_ylabel('')

plt.tight_layout()
# %%

f, axes = plt.subplots(ncols=2, figsize=(12, 6), sharex=True, sharey=True)

axes[0].set_title('IRASA')
sns.regplot(
    df_cmb_cut.query('param_type == "irasa"'),
    x='Error (Exponent)',
    y='Knee Frequency (Hz)',
    # hue='param_type',
    ax=axes[0],
)

axes[1].set_title('No IRASA')
sns.regplot(
    df_cmb_cut.query('param_type == "norasa"'),
    x='Error (Exponent)',
    y='Knee Frequency (Hz)',
    # hue='param_type',
    ax=axes[1],
)
# %%
f, axes = plt.subplots(ncols=2, figsize=(12, 6), sharex=True, sharey=True)

axes[0].set_title('IRASA')
sns.regplot(
    df_cmb_cut.query('param_type == "irasa"'),
    x='Error (Knee Frequency (Hz))',
    y='Knee Frequency (Hz)',
    # hue='param_type',
    ax=axes[0],
)
axes[1].set_title('No IRASA')
sns.regplot(
    df_cmb_cut.query('param_type == "norasa"'),
    x='Error (Knee Frequency (Hz))',
    y='Knee Frequency (Hz)',
    # hue='param_type',
    ax=axes[1],
)
# %%



# %%
from neurodsp.sim import set_random_seed
from neurodsp.sim import sim_powerlaw, sim_oscillation
from neurodsp.utils import create_times
from neurodsp.plts import plot_timefrequency  #

from neurodsp.timefrequency import compute_wavelet_transform
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns

sns.set_style('ticks')
sns.set_context('talk')

set_random_seed(42)

from pyrasa.irasa import irasa_sprint

# %%
# Set some general settings, to be used across all simulations
fs = 500
n_seconds = 15
duration = 4
overlap = 0.5

# Create a times vector for the simulations
times = create_times(n_seconds, fs)

# %%
alpha = sim_oscillation(n_seconds=0.5, fs=fs, freq=10)
no_alpha = np.zeros(len(alpha))
beta = sim_oscillation(n_seconds=0.5, fs=fs, freq=25)
no_beta = np.zeros(len(beta))

exp_1 = sim_powerlaw(n_seconds=2.5, fs=fs, exponent=-1)
exp_2 = sim_powerlaw(n_seconds=2.5, fs=fs, exponent=-2)

# %%
alphas = np.concatenate([no_alpha, alpha, no_alpha, alpha, no_alpha])
betas = np.concatenate([beta, no_beta, beta, no_beta, beta])

sim_ts = np.concatenate(
    [
        exp_1 + alphas,
        exp_1 + alphas + betas,
        exp_1 + betas,
        exp_2 + alphas,
        exp_2 + alphas + betas,
        exp_2 + betas,
    ]
)
# %%
plt.plot(times, sim_ts)
# %%
freqs = np.arange(1, 50, 0.5)

# %%
mwt = compute_wavelet_transform(
    sim_ts,
    fs=fs,
    freqs=freqs,
    n_cycles=11,
)
# %%
plot_timefrequency(times, freqs, mwt, vmin=0)

# %%
sgramm_ap, sgramm_p, freqs_ir, times_ir = irasa_sprint(
    sim_ts[np.newaxis, :], fs=fs, band=(1, 100), freq_res=0.5, smooth=False, n_avgs=[3, 7, 11]
)


# %% now extract the aperiodic features
from pyrasa.utils.aperiodic_utils import compute_slope_sprint

df_aps, df_gof = compute_slope_sprint(sgramm_ap, freqs=freqs_ir, times=times_ir, fit_func='fixed')


plot_timefrequency(times_ir, freqs_ir, np.squeeze(sgramm_ap), vmin=0)


# %%
ave = df_gof['r_squared'].mean()
std = df_gof['r_squared'].std()
print(f'Average goodness of fit {ave} and deviation is {std}')


# %%
ave = df_gof.query('time < 7')['r_squared'].mean()
std = df_gof.query('time < 7')['r_squared'].std()
print(f'Average goodness of fit {ave} and deviation is {std}')


# %%
ave = df_gof.query('time > 7')['r_squared'].mean()
std = df_gof.query('time > 7')['r_squared'].std()
print(f'Average goodness of fit {ave} and deviation is {std}')


# %%
f, ax = plt.subplots(nrows=2, figsize=(8, 4))

ax[0].plot(df_aps['time'], df_aps['Exponent'])
ax[0].set_ylabel('Exponent')
ax[0].set_xlabel('time (s)')
ax[1].plot(df_aps['time'], df_gof['r_squared'])
ax[1].set_ylabel('R2')
ax[1].set_xlabel('time (s)')

f.tight_layout()

# %%
# %%
from pyrasa.utils.peak_utils import get_peak_params_sprint

df_peaks = get_peak_params_sprint(sgramm_p, freqs=freqs_ir, times=times_ir, min_peak_height=0.01)


plot_timefrequency(times_ir, freqs_ir, np.squeeze(sgramm_p), vmin=0)

# %% Plot peak results
f, ax = plt.subplots(nrows=3, figsize=(12, 4))

for ix, cur_key in enumerate(['cf', 'pw', 'bw']):
    ax[ix].plot(df_peaks['time'], df_peaks[cur_key])
    ax[ix].set_ylabel(cur_key)
    ax[ix].set_xlabel('time (s)')

# %%



"""Simulation settings."""

import numpy as np

###################################################################################################
## Define general simulation settings

# Set the number of instances to run (within method simulations)
N_SIMS = 50
N_SIMS2 = 10

# Set the number of samples to run (between method comparisons)
N_SAMPLES = 1000

# General simulation settings
N_SECONDS = 30
FS = 250
FS2 = 500

###################################################################################################
## Define default parameter values

# Define aperiodic parameters
EXP = -1
EXP1 = 0
EXP2 = -2
KNEE = 100

# Define additional parameters related to aperiodic components
F_RANGE = (0.5, None)

# Define periodic parameters
FREQ = 10
BW = 1.5
HEIGHT = 1.5

# Define parameters for combined signals
COMP_VARS = [1, 0.5]

###################################################################################################
## Define parameter ranges

# Aperiodic parameters - ranges
EXPS = np.round(np.arange(-3, 0.5, 0.5), 1)
# TSCALES = np.round(np.array([0.005, 0.015, 0.030, 0.050, 0.075]))
KNEES = np.array([25, 100, 400, 900, 1600])

# Periodic parameter - ranges
FREQS = np.arange(5, 36, 1)
POWERS = np.round(np.arange(0, 2.1, 0.1), 1)
CVARS = [[1, val] for val in POWERS]
BWS = np.round(np.arange(0.5, 3.5, 0.5), 1)

# Burst related parameters
BPROBS = np.round(np.arange(0.2, 0.8, 0.1), 1)



import numpy as np
import pytest
import scipy.signal as dsp

from pyrasa import irasa, irasa_sprint

from .settings import EXP_KNEE_COMBO, FS, KNEE_TOLERANCE, MIN_CORR_PSD_CMB, OSC_FREQ, TOLERANCE

# Estimate periodic and aperiodic components with IRASA
# These tests should cover the basic functionality of the IRASA workflow


# knee model
@pytest.mark.parametrize('exponent, knee', EXP_KNEE_COMBO, scope='session')
@pytest.mark.parametrize('fs', FS, scope='session')
def test_irasa_knee_peakless(load_knee_aperiodic_signal, fs, exponent, knee):
    f_range = [0.1, 100]
    irasa_out = irasa(load_knee_aperiodic_signal, fs, f_range, psd_kwargs={'nperseg': 4 * fs})
    # test the shape of the output
    assert irasa_out.freqs.shape[0] == irasa_out.aperiodic.shape[1] == irasa_out.periodic.shape[1]
    freqs_psd, psd = dsp.welch(load_knee_aperiodic_signal, fs, nperseg=int(4 * fs))
    psd_cmb = irasa_out.aperiodic[0, :] + irasa_out.periodic[0, :]
    freq_logical = np.logical_and(freqs_psd >= f_range[0], freqs_psd <= f_range[1])
    r = np.corrcoef(psd[freq_logical], psd_cmb)[0, 1]
    assert r > MIN_CORR_PSD_CMB
    slope_fit_k = irasa_out.fit_aperiodic_model(fit_func='knee')
    slope_fit_f = irasa_out.fit_aperiodic_model(fit_func='fixed')
    # test whether we can get the first exponent correctly
    assert bool(np.isclose(slope_fit_k.aperiodic_params['Exponent_1'][0], 0, atol=TOLERANCE))
    # test whether we can get the second exponent correctly
    assert bool(
        np.isclose(
            slope_fit_k.aperiodic_params['Exponent_2'][0],
            np.abs(exponent),
            atol=TOLERANCE,
        )
    )
    # test whether we can get the knee correctly
    knee_hat = slope_fit_k.aperiodic_params['Knee'][0] ** (
        1 / (2 * slope_fit_k.aperiodic_params['Exponent_1'][0] + slope_fit_k.aperiodic_params['Exponent_2'][0])
    )
    knee_real = knee ** (1 / np.abs(exponent))
    assert bool(np.isclose(knee_hat, knee_real, atol=KNEE_TOLERANCE))
    # test bic/aic -> should be better for knee
    assert slope_fit_k.gof['AIC'][0] < slope_fit_f.gof['AIC'][0]
    assert slope_fit_k.gof['BIC_adj.'][0] < slope_fit_f.gof['BIC_adj.'][0]


# knee model
@pytest.mark.parametrize('exponent, knee', EXP_KNEE_COMBO, scope='session')
@pytest.mark.parametrize('fs', FS, scope='session')
@pytest.mark.parametrize('osc_freq', OSC_FREQ, scope='session')
def test_irasa_knee_cmb(load_knee_cmb_signal, fs, exponent, knee, osc_freq):
    f_range = [0.1, 100]
    irasa_out = irasa(load_knee_cmb_signal, fs, f_range, psd_kwargs={'nperseg': 4 * fs})
    # test the shape of the output
    assert irasa_out.freqs.shape[0] == irasa_out.aperiodic.shape[1] == irasa_out.periodic.shape[1]
    freqs_psd, psd = dsp.welch(load_knee_cmb_signal, fs, nperseg=int(4 * fs))
    psd_cmb = irasa_out.aperiodic[0, :] + irasa_out.periodic[0, :]
    freq_logical = np.logical_and(freqs_psd >= f_range[0], freqs_psd <= f_range[1])
    r = np.corrcoef(psd[freq_logical], psd_cmb)[0, 1]
    assert r > MIN_CORR_PSD_CMB
    slope_fit_k = irasa_out.fit_aperiodic_model(fit_func='knee')
    slope_fit_f = irasa_out.fit_aperiodic_model(fit_func='fixed')
    # test whether we can get the first exponent correctly
    assert bool(np.isclose(slope_fit_k.aperiodic_params['Exponent_1'][0], 0, atol=TOLERANCE))
    # test whether we can get the second exponent correctly
    assert bool(
        np.isclose(
            slope_fit_k.aperiodic_params['Exponent_2'][0],
            np.abs(exponent),
            atol=TOLERANCE,
        )
    )
    # test whether we can get the knee correctly
    knee_hat = slope_fit_k.aperiodic_params['Knee'][0] ** (
        1 / (2 * slope_fit_k.aperiodic_params['Exponent_1'][0] + slope_fit_k.aperiodic_params['Exponent_2'][0])
    )
    knee_real = knee ** (1 / np.abs(exponent))
    assert bool(np.isclose(knee_hat, knee_real, atol=KNEE_TOLERANCE))
    # test bic/aic -> should be better for knee
    assert slope_fit_k.gof['AIC'][0] < slope_fit_f.gof['AIC'][0]
    assert slope_fit_k.gof['BIC_adj.'][0] < slope_fit_f.gof['BIC_adj.'][0]
    # test whether we can reconstruct the peak frequency correctly
    pe_params = irasa_out.get_peaks()
    assert bool(np.isclose(np.round(pe_params['cf'], 0), osc_freq))


@pytest.mark.parametrize('exponent, knee', [(-1.5, 1000)], scope='session')
@pytest.mark.parametrize('fs', [1000], scope='session')
@pytest.mark.parametrize('osc_freq', [10], scope='session')
def test_aperiodic_error(load_knee_cmb_signal, fs, exponent, knee, osc_freq):
    duration = 4
    overlap = 0.5
    irasa_out = irasa(
        load_knee_cmb_signal,
        fs=fs,
        band=(0.1, 50),
        psd_kwargs={'nperseg': duration * fs, 'noverlap': duration * fs * overlap},
        hset_info=(1, 2.0, 0.05),
    )

    irasa_out_bad = irasa(
        load_knee_cmb_signal,
        fs=fs,
        band=(0.1, 50),
        psd_kwargs={'nperseg': duration * fs, 'noverlap': duration * fs * overlap},
        hset_info=(1, 8.0, 0.05),
    )

    assert np.mean(irasa_out.get_aperiodic_error()) < np.mean(irasa_out_bad.get_aperiodic_error())


@pytest.mark.parametrize('fs', [1000], scope='session')
@pytest.mark.parametrize('exponent_1', [-0], scope='session')
@pytest.mark.parametrize('exponent_2', [-2], scope='session')
def test_aperiodic_error_tf(ts4sprint_knee, fs, exponent_1, exponent_2):
    irasa_out = irasa_sprint(
        ts4sprint_knee,
        fs=fs,
        band=(0.1, 50),
        overlap_fraction=0.95,
        win_duration=0.5,
        hset_info=(1, 2.0, 0.05),
    )

    irasa_out_bad = irasa_sprint(
        ts4sprint_knee,
        fs=fs,
        band=(0.1, 50),
        overlap_fraction=0.95,
        win_duration=0.5,
        hset_info=(1, 8.0, 0.05),
    )

    kwargs = {
        'cut_spectrum': (1, 40),
        'smooth': True,
        'smoothing_window': 3,
        'min_peak_height': 0.01,
        'peak_width_limits': (0.5, 12),
    }

    assert np.mean(irasa_out.get_aperiodic_error(peak_kwargs=kwargs)) < np.mean(
        irasa_out_bad.get_aperiodic_error(peak_kwargs=kwargs)
    )



# %%
import joblib
from pathlib import Path
import numpy as np
import pandas as pd

import matplotlib.pyplot as plt
import seaborn as sns

import matplotlib as mpl

new_rc_params = {'text.usetex': False, 'svg.fonttype': 'none'}
mpl.rcParams.update(new_rc_params)

sns.set_style('ticks')
sns.set_context('poster')

import mne


# %%
data_l, data_l_ap = [], []
for i in np.arange(2, 11):
    base_p = f'/home/schmidtfa/git/pyrasa/paper/data/eo_ec/{i}/'
    cur_ps = list(Path(base_p).glob('*.dat'))

    cur_data = pd.concat([joblib.load(cur)['aperiodics']['fixed'] for cur in cur_ps])
    data_l_ap.append(cur_data)

    cur_data = pd.concat([joblib.load(cur)['periodics']['alpha'] for cur in cur_ps])
    data_l.append(cur_data)
# %%
all_df_p = pd.concat(data_l)
# %%
df_comp = all_df_p.groupby(['s_id', 'condition'])[['cf', 'bw', 'pw']].mean().reset_index()
# %%
f, axes = plt.subplots(ncols=3, figsize=(12, 4))

swarm_kwargs = {'size': 10, 'alpha': 0.5}
point_kwargs = {'markers': '_', 'markersize': 20}

pretty_labels = ['Center Frequency (Hz)', 'Bandwidth (Hz)', 'Power']

for ix, (cond, gl) in enumerate(zip(['cf', 'bw', 'pw'], pretty_labels)):
    sns.stripplot(data=df_comp, x='condition', y=cond, hue='condition', ax=axes[ix], **swarm_kwargs)
    sns.pointplot(data=df_comp, x='condition', y=cond, hue='condition', ax=axes[ix], **point_kwargs)
    axes[ix].set_xlabel('')
    axes[ix].set_ylabel(gl)

sns.despine()
f.tight_layout()

f.savefig('../results/periodic_eo_ec.svg')

# %%
all_df_ap = pd.concat(data_l_ap)
all_df_ap.drop(columns='fit_type', inplace=True)
# %%
df_comp = all_df_ap.groupby(['s_id', 'condition'])[['Offset', 'Exponent']].mean().reset_index()
# %%
f, axes = plt.subplots(ncols=2, figsize=(8, 4))

swarm_kwargs = {'size': 10, 'alpha': 0.5}
point_kwargs = {'markers': '_', 'markersize': 20}

for ix, cond in enumerate(['Offset', 'Exponent']):
    sns.stripplot(data=df_comp, x='condition', y=cond, hue='condition', ax=axes[ix], **swarm_kwargs)
    sns.pointplot(data=df_comp, x='condition', y=cond, hue='condition', ax=axes[ix], **point_kwargs)
    axes[ix].set_xlabel('')

sns.despine()
f.tight_layout()

f.savefig('../results/aperiodic_eo_ec.svg')
# %%

df_cmb = (
    all_df_p.merge(all_df_ap, on=['s_id', 'condition', 'ch_name'])
    .groupby(['s_id', 'condition'])[['Offset', 'Exponent', 'cf', 'bw', 'pw']]
    .mean()
    .reset_index()
)
corrs = df_cmb.drop(columns='s_id').groupby('condition').corr('spearman')

ec_corr = corrs.loc['EC']
eo_corr = corrs.loc['EO']

# %%
f, ax = plt.subplots(figsize=(16, 6), ncols=2)
mask = np.triu(np.ones_like(ec_corr, dtype=bool))
sns.heatmap(ec_corr, annot=True, vmin=-1, vmax=1, cmap='RdBu_r', ax=ax[0])
sns.heatmap(eo_corr, annot=True, vmin=-1, vmax=1, cmap='RdBu_r', ax=ax[1])

# %%
df_cmb.query('condition == "EC"')[['Offset', 'Exponent', 'cf', 'bw', 'pw']].corr()
# %%
ch_names = [
    'Fp1',
    'AF7',
    'AF3',
    'F1',
    'F3',
    'F5',
    'F7',
    'FT7',
    'FC5',
    'FC3',
    'FC1',
    'C1',
    'C3',
    'C5',
    'T7',
    'TP7',
    'CP5',
    'CP3',
    'CP1',
    'P1',
    'P3',
    'P5',
    'P7',
    'P9',
    'PO7',
    'PO3',
    'O1',
    'Iz',
    'Oz',
    'POz',
    'Pz',
    'CPz',
    'Fpz',
    'Fp2',
    'AF8',
    'AF4',
    'Afz',
    'Fz',
    'F2',
    'F4',
    'F6',
    'F8',
    'FT8',
    'FC6',
    'FC4',
    'FC2',
    'FCz',
    'Cz',
    'C2',
    'C4',
    'C6',
    'T8',
    'TP8',
    'CP6',
    'CP4',
    'CP2',
    'P2',
    'P4',
    'P6',
    'P8',
    'P10',
    'PO8',
    'PO4',
    'O26',
]

# %%
df_conds = all_df_p.groupby(['condition', 'ch_name']).median().reset_index()
ec = df_conds.query('condition == "EC"')
eo = df_conds.query('condition == "EO"')

ec = ec.set_index('ch_name').reindex(ch_names)
eo = eo.set_index('ch_name').reindex(ch_names)
# %%
montage = mne.channels.make_standard_montage(kind='biosemi64')
info = mne.create_info(ch_names=montage.ch_names, sfreq=264, ch_types='eeg')

eo_im = mne.EvokedArray(eo['pw'].to_numpy()[:, np.newaxis], info=info)
eo_im.set_montage('biosemi64')

ec_im = mne.EvokedArray(ec['pw'].to_numpy()[:, np.newaxis], info=info)
ec_im.set_montage('biosemi64')

ec_eo_im = mne.EvokedArray((ec['pw'].to_numpy() - eo['pw'].to_numpy())[:, np.newaxis], info=info)
ec_eo_im.set_montage('biosemi64')

# %%
min, max = eo['pw'].to_numpy().mean(), ec['pw'].to_numpy().max()

sns.set_context('talk')
f, ax = plt.subplots(figsize=(4, 4), nrows=2)
eo_im.plot_topomap(
    ch_type='eeg',
    times=0,
    colorbar=False,  # vlim=(min, max),
    cmap='Reds',
    axes=ax[0],
)

ec_im.plot_topomap(ch_type='eeg', times=0, colorbar=False, vlim=(min, max), cmap='Reds', axes=ax[1])
# %%
f, ax = plt.subplots(figsize=(4, 4))
sns.set_context('talk')
ec_eo_im.plot_topomap(ch_type='eeg', times=0, colorbar=False, cmap='Reds', axes=ax)
# %%



"""Output Class of pyrasa.irasa_sprint"""

import numpy as np
import pandas as pd
from attrs import define

from pyrasa.utils.aperiodic_utils import compute_aperiodic_model_sprint
from pyrasa.utils.fit_funcs import AbstractFitFun
from pyrasa.utils.peak_utils import get_peak_params_sprint
from pyrasa.utils.types import AperiodicFit

min_ndim = 2


@define
class IrasaTfSpectrum:
    freqs: np.ndarray
    time: np.ndarray
    raw_spectrum: np.ndarray
    aperiodic: np.ndarray
    periodic: np.ndarray
    ch_names: np.ndarray | None

    def fit_aperiodic_model(
        self,
        fit_func: str | type[AbstractFitFun] = 'fixed',
        scale: bool = False,
        fit_bounds: tuple[float, float] | None = None,
    ) -> AperiodicFit:
        """
        Extracts aperiodic parameters from the aperiodic spectrogram using scipy's curve fitting
        function.

        This function computes aperiodic parameters for each time point in the spectrogram by applying either one of
        two different curve fitting functions (`fixed` or `knee`) or a custom function specified by user to the data.
        See examples custom_fit_functions.ipynb. The parameters, along with the goodness of
        fit for each time point, are returned in a concatenated format.

        Parameters
        ----------
        fit_func : str or type[AbstractFitFun], optional
            The fitting function to use. Can be "fixed" for a linear fit or "knee" for a fit that includes a
            knee (bend) in the spectrum or a class that is inherited from AbstractFitFun. The default is 'fixed'..
        scale : bool, optional
            Whether to scale the data to improve fitting accuracy. This is useful when fitting a knee in cases where
            power values are very small, leading to numerical precision issues. Default is False.
        fit_bounds : tuple[float, float] or None, optional
            Tuple specifying the lower and upper frequency bounds for the fit function. If None, the entire frequency
            range is used. Otherwise, the spectrum is cropped to the specified bounds before fitting. Default is None.

        Returns
        -------
        AperiodicFit
            An object containing two pandas DataFrames:
                - aperiodic_params : pd.DataFrame
                    A DataFrame containing the aperiodic parameters (e.g., center frequency, bandwidth, peak height)
                    for each channel and each time point.
                - gof : pd.DataFrame
                    A DataFrame containing the goodness of fit metrics for each channel and each time point.

        Notes
        -----
        This function iterates over each time point in the provided spectrogram to extract aperiodic parameters
        using the specified fit function. It leverages the `compute_aperiodic_model` function for individual fits
        at each time point, then combines the results across all time points into comprehensive DataFrames.

        The `fit_bounds` parameter allows for frequency range restrictions during fitting, which can help in focusing
        the analysis on a particular frequency band of interest.

        Scaling the data using the `scale` parameter can be particularly important when dealing with very small power
        values that might lead to poor fitting due to numerical precision limitations.

        """

        return compute_aperiodic_model_sprint(
            aperiodic_spectrum=self.aperiodic[np.newaxis, :, :] if self.aperiodic.ndim == min_ndim else self.aperiodic,
            freqs=self.freqs,
            times=self.time,
            ch_names=self.ch_names,
            scale=scale,
            fit_func=fit_func,
            fit_bounds=fit_bounds,
        )

    def get_peaks(
        self,
        smooth: bool = True,
        smoothing_window: float | int = 1,
        cut_spectrum: tuple[float, float] | None = None,
        peak_threshold: float = 2.5,
        min_peak_height: float = 0.0,
        polyorder: int = 1,
        peak_width_limits: tuple[float, float] = (0.5, 12),
    ) -> pd.DataFrame:
        """
        Extracts peak parameters from a periodic spectrogram obtained via IRASA.

        This method processes a time-resolved periodic spectrum to identify and extract peak parameters such as
        center frequency (cf), bandwidth (bw), and peak height (pw) for each time point. It applies smoothing,
        peak detection, and thresholding according to user-defined parameters
        (see get_peak_params for additional Information).

        Parameters
        ----------
        smooth : bool, optional
            Whether to smooth the spectrum before peak extraction. Smoothing can help in reducing noise and better
            identifying peaks. Default is True.
        smoothing_window : int or float, optional
            The size of the smoothing window in Hz, passed to the Savitzky-Golay filter. Default is 2 Hz.
        polyorder : int, optional
            The polynomial order for the Savitzky-Golay filter used in smoothing. The polynomial order must be less
            than the window length. Default is 1.
        cut_spectrum : tuple of (float, float) or None, optional
            Tuple specifying the frequency range (lower_bound, upper_bound) to which the spectrum should be cut before
            peak extraction. If None, the full frequency range is used. Default is (1, 40).
        peak_threshold : int or float, optional
            Relative threshold for detecting peaks, defined as a multiple of the standard deviation of the filtered
            spectrum. Default is 1.
        min_peak_height : float, optional
            The minimum peak height (in absolute units of the power spectrum) required for a peak to be recognized.
            This can be useful for filtering out noise or insignificant peaks, especially when a "knee" is present
            in the data. Default is 0.01.
        peak_width_limits : tuple of (float, float), optional
            The lower and upper bounds for peak widths, in Hz. This helps in constraining the peak detection to
            meaningful features. Default is (0.5, 12).

        Returns
        -------
        pd.DataFrame
            A DataFrame containing the detected peak parameters for each channel and time point. The DataFrame
            includes the following columns:
            - 'ch_name': Channel name
            - 'cf': Center frequency of the peak
            - 'bw': Bandwidth of the peak
            - 'pw': Peak height (power)
            - 'time': Corresponding time point for the peak

        Notes
        -----
        This function iteratively processes each time point in the spectrogram, applying the `get_peak_params`
        function to extract peak parameters at each time point. The resulting peak parameters are combined into
        a single DataFrame.

        The function is particularly useful for analyzing time-varying spectral features, such as in dynamic or
        non-stationary M/EEG data, where peaks may shift in frequency, bandwidth, or amplitude over time.

        """

        return get_peak_params_sprint(
            periodic_spectrum=self.periodic[np.newaxis, :, :] if self.periodic.ndim == min_ndim else self.periodic,
            freqs=self.freqs,
            times=self.time,
            ch_names=self.ch_names,
            smooth=smooth,
            smoothing_window=smoothing_window,
            cut_spectrum=cut_spectrum,
            peak_threshold=peak_threshold,
            min_peak_height=min_peak_height,
            polyorder=polyorder,
            peak_width_limits=peak_width_limits,
        )

    def get_aperiodic_error(self, peak_kwargs: None | dict = None) -> np.ndarray:
        """
        Computes the frequency resolved error of the aperiodic spectrum.

        This method first computes the absolute of the periodic spectrum and subsequently zeroes out
        any peaks in the spectrum that are potentially "oscillations", yielding the residual error of the aperiodic
        spectrum as a function of frequency.
        This can be useful when trying to optimize hyperparameters such as the hset.

        peak_kwargs : dict
            A dictionary containing keyword arguments that are passed on to the peak finding method 'get_peaks'

        Returns
        -------
            np.ndarray
                A numpy array containing the frequency resolved squared error of the aperiodic
                spectrum extracted using irasa


        Notes
        -----
        While not strictly necessary, setting peak_kwargs is highly recommended.
        The reason for this is that through up-/downsampling and averaging "broadband"
        parameters such as spectral knees can bleed in the periodic spectrum and could be wrongfully
        interpreted as oscillations. This can be avoided by e.g. explicitely setting `min_peak_height`.
        A good way of making a decision for the periodic parameters is to base it on the settings
        used in peak detection.

        """

        if peak_kwargs is None:
            peak_kwargs = {}

        # get absolute periodic spectrum & zero-out peaks
        freqs = self.freqs
        peaks = self.get_peaks(**peak_kwargs)
        peak_times = peaks['time'].unique()
        ch_names = peaks['ch_name'].unique()

        valid_peak_times = [cur_t in peak_times for cur_t in self.time]
        aperiodic_error = np.abs(self.periodic)
        aperiodic_error_cut = aperiodic_error[:, :, valid_peak_times]

        aperiodic_errors_ch = []
        for c_ix, ch in enumerate(ch_names):
            cur_ch_ape = aperiodic_error_cut[c_ix, :, :]
            cur_peak_ch = peaks.query(f'ch_name == "{ch}"')

            aperiodic_errors_t = []
            for t_ix, cur_t in enumerate(peak_times):
                cur_t_ape = cur_ch_ape[:, t_ix]
                cur_t_ch = cur_peak_ch.query(f'time == "{cur_t}"')

                for _, peak in cur_t_ch.iterrows():
                    cur_upper = peak['cf'] + peak['bw']
                    cur_lower = peak['cf'] - peak['bw']

                    freq_mask = np.logical_and(freqs < cur_upper, freqs > cur_lower)

                    cur_t_ape[freq_mask] = 0

                aperiodic_errors_t.append(cur_t_ape)

            aperiodic_errors_ch.append(np.array(aperiodic_errors_t).T)

        return np.array(aperiodic_errors_ch)



# %%
from neurodsp.sim import sim_powerlaw, sim_peak_oscillation
import numpy as np
import scipy.signal as dsp

import matplotlib.pyplot as plt
import seaborn as sns

sns.set_style('ticks')
sns.set_context('poster')

from pyrasa.irasa import irasa
from pyrasa.utils.peak_utils import get_peak_params
from pyrasa.utils.aperiodic_utils import compute_slope

# %%
n_secs = 4
fs = 1000

sig_ap = sim_powerlaw(n_seconds=n_secs, fs=fs, exponent=-1)
sig = sim_peak_oscillation(sig_ap=sig_ap, fs=fs, freq=10, bw=3, height=3)

sig_cmb = np.tile(np.concatenate([sig_ap, sig]), 20)

# %%
# osc = sim_oscillation(n_seconds=n_secs, fs=fs, freq=10) + sig_ap

# %%
freq, psd = dsp.welch(x=sig_cmb, fs=fs, nperseg=4 * fs, noverlap=2 * fs)
# freq, psd_osc = dsp.welch(x=osc, fs=fs, nperseg=4*fs, noverlap=2*fs)

f_logical = np.logical_and(freq > 1, freq < 100)

# f_logical = np.logical_and(freq > 8, freq < 12)
f, ax = plt.subplots(figsize=(4, 4))
ax.loglog(freq[f_logical], psd[f_logical])
# ax.loglog(freq[f_logical], psd_osc[f_logical])
# %%
freq_rasa, aperiodic, periodic = irasa(
    sig_cmb,
    fs=fs,
    band=(1, 100),
    kwargs_psd={'nperseg': 4 * fs, 'noverlap': 2 * fs, 'average': 'median'},
    hset_info=(1.0, 4.0, 0.25),
)

# %%
plt.loglog(freq_rasa, aperiodic[0, :])
# %%
df_ap, df_gof = compute_slope(aperiodic, freqs=freq_rasa, fit_func='fixed')

df_ap
# %%
df_gof
# %%

f_logical = np.logical_and(freq_rasa > 1, freq_rasa < 20)
plt.plot(freq_rasa[f_logical], periodic[0, f_logical] * 10)
# f_logical = np.logical_and(freq > 8, freq < 12)
# plt.plot(freq[f_logical], psd[f_logical])

get_peak_params(periodic, freqs=freq_rasa, min_peak_height=0.1)

# %%



import mne
import numpy as np
import pytest
from mne.datasets import sample
from neurodsp.sim import sim_combined, sim_knee, sim_oscillation, sim_powerlaw
from neurodsp.utils.sim import set_random_seed

from .settings import N_SECONDS

# def pytest_configure(config):
set_random_seed(42)


@pytest.fixture(scope='session')
def combined_signal(exponent, osc_freq, fs):
    components = {'sim_powerlaw': {'exponent': exponent}, 'sim_oscillation': {'freq': osc_freq}}
    yield sim_combined(n_seconds=N_SECONDS, fs=fs, components=components)


@pytest.fixture(scope='session')
def fixed_aperiodic_signal(exponent, fs):
    yield sim_powerlaw(n_seconds=N_SECONDS, fs=fs, exponent=exponent)


@pytest.fixture(scope='session')
def knee_aperiodic_signal(exponent, fs, knee_freq):
    yield sim_knee(n_seconds=N_SECONDS, fs=fs, exponent1=0, exponent2=exponent, knee=knee_freq ** np.abs(exponent))  # noqa: E501


@pytest.fixture(scope='session')
def load_knee_aperiodic_signal(exponent, fs, knee):
    # % generate and save knee
    # knee = knee ** np.abs(exponent)
    knee_sim = sim_knee(n_seconds=N_SECONDS, fs=fs, exponent1=0, exponent2=exponent, knee=knee)
    yield knee_sim
    # base_dir = 'tests/test_data/knee_data/'
    # yield np.load(
    #     base_dir + f'knee_sim__fs_{fs}__exp1_0__exp2_{exponent}_knee_{knee}_.npy',
    # )  # allow_pickle=True)


@pytest.fixture(scope='session')
def load_knee_cmb_signal(exponent, fs, knee, osc_freq):
    # knee = knee ** np.abs(exponent)
    components = {
        'sim_knee': {'exponent1': 0, 'exponent2': exponent, 'knee': knee},
        'sim_oscillation': {'freq': osc_freq},
    }
    cmb_sim = sim_combined(n_seconds=N_SECONDS, fs=fs, components=components)
    yield cmb_sim
    # base_dir = 'tests/test_data/knee_osc_data/'
    # yield np.load(
    #     base_dir + f'cmb_sim__fs_{fs}__exp1_0__exp2_{exponent}_knee_{knee}__osc_freq_{osc_freq}_.npy',
    # )  # allow_pickle=True)
    # noqa: E501


@pytest.fixture(scope='session')
def oscillation(osc_freq, fs):
    yield sim_oscillation(n_seconds=N_SECONDS, fs=fs, freq=osc_freq)


@pytest.fixture(scope='session')
def ts4sprint(fs, exponent_1, exponent_2):
    alpha = sim_oscillation(n_seconds=0.5, fs=fs, freq=10)
    no_alpha = np.zeros(len(alpha))
    beta = sim_oscillation(n_seconds=0.5, fs=fs, freq=25)
    no_beta = np.zeros(len(beta))

    exp_1 = sim_powerlaw(n_seconds=2.5, fs=fs, exponent=exponent_1)
    exp_2 = sim_powerlaw(n_seconds=2.5, fs=fs, exponent=exponent_2)

    # %%
    alphas = np.concatenate([no_alpha, alpha, no_alpha, alpha, no_alpha])
    betas = np.concatenate([beta, no_beta, beta, no_beta, beta])

    sim_ts = np.concatenate(
        [
            exp_1 + alphas,
            exp_1 + alphas + betas,
            exp_1 + betas,
            exp_2 + alphas,
            exp_2 + alphas + betas,
            exp_2 + betas,
        ]
    )
    yield sim_ts


@pytest.fixture(scope='session')
def ts4sprint_knee(fs, exponent_1, exponent_2):
    alpha = sim_oscillation(n_seconds=0.5, fs=fs, freq=10)
    no_alpha = np.zeros(len(alpha))
    beta = sim_oscillation(n_seconds=0.5, fs=fs, freq=25)
    no_beta = np.zeros(len(beta))

    knee1 = 20 ** np.abs(exponent_1)
    knee2 = 20 ** np.abs(exponent_2)
    exp_1 = sim_knee(n_seconds=2.5, fs=fs, exponent1=0, exponent2=exponent_1, knee=knee1)
    exp_2 = sim_knee(n_seconds=2.5, fs=fs, exponent1=0, exponent2=exponent_2, knee=knee2)

    # %%
    alphas = np.concatenate([no_alpha, alpha, no_alpha, alpha, no_alpha])
    betas = np.concatenate([beta, no_beta, beta, no_beta, beta])

    sim_ts = np.concatenate(
        [
            exp_1 + alphas,
            exp_1 + alphas + betas,
            exp_1 + betas,
            exp_2 + alphas,
            exp_2 + alphas + betas,
            exp_2 + betas,
        ]
    )
    yield sim_ts


@pytest.fixture(scope='session')
def gen_mne_data_raw():
    data_path = sample.data_path()

    meg_path = data_path / 'MEG' / 'sample'
    raw_fname = meg_path / 'sample_audvis_raw.fif'
    event_fname = meg_path / 'sample_audvis_filt-0-40_raw-eve.fif'
    events = mne.read_events(event_fname)

    raw = mne.io.read_raw_fif(raw_fname)
    picks = mne.pick_types(raw.info, meg='mag', eeg=False, stim=False, eog=False, exclude='bads')
    raw.pick(picks)

    # % now lets check-out the events
    event_id = {
        'Auditory/Left': 1,
        'Auditory/Right': 2,
        'Visual/Left': 3,
        'Visual/Right': 4,
    }
    tmin = -0.2
    tmax = 0.5

    # Load real data as the template
    epochs = mne.Epochs(
        raw,
        events,
        event_id,
        tmin,
        tmax,
        baseline=None,
        preload=True,
        verbose=False,
    )

    yield raw, epochs



"""Utilities for extracting peak parameters."""

from collections.abc import Iterable

import numpy as np
import pandas as pd
import scipy.signal as dsp


# %% find peaks irasa style
def get_peak_params(
    periodic_spectrum: np.ndarray,
    freqs: np.ndarray,
    ch_names: Iterable | None = None,
    smooth: bool = True,
    smoothing_window: int | float = 1,
    polyorder: int = 1,
    cut_spectrum: tuple[float, float] | None = None,
    peak_threshold: float = 1.0,
    min_peak_height: float = 0.01,
    peak_width_limits: tuple[float, float] = (0.5, 12.0),
) -> pd.DataFrame:
    """
    Extracts peak parameters from the periodic spectrum obtained via IRASA.

    This function identifies and extracts peak parameters such as center frequency (cf), bandwidth (bw),
    and peak height (pw) from a periodic spectrum using scipy's find_peaks function.
    The spectrum can be optionally smoothed prior peak detection.

    Parameters
    ----------
    periodic_spectrum : np.ndarray
        1D or 2D array containing power values of the periodic spectrum (shape: [Channels, Frequencies]
        or [Frequencies]).
    freqs : np.ndarray
        1D array containing frequency values corresponding to the periodic spectrum.
    ch_names : Iterable or None, optional
        List of channel names corresponding to the periodic spectrum. If None, channels are labeled numerically.
        Default is None.
    smooth : bool, optional
        Whether to smooth the spectrum before peak extraction. Smoothing can help in reducing noise and
        better identifying peaks. Default is True.
    smoothing_window : int or float, optional
        The size of the smoothing window in Hz, passed to the Savitzky-Golay filter. Default is 1 Hz.
    polyorder : int, optional
        The polynomial order for the Savitzky-Golay filter used in smoothing. The polynomial order must be
        less than the window length. Default is 1.
    cut_spectrum : tuple of (float, float) or None, optional
        Tuple specifying the frequency range (lower_bound, upper_bound) to which the spectrum should be cut
        before peak extraction. If None, peaks are detected across the full frequency range. Default is None.
    peak_threshold : float, optional
        Relative threshold for detecting peaks, defined as a multiple of the standard deviation of the
        filtered spectrum. Default is 1.0.
    min_peak_height : float, optional
        The minimum peak height (in absolute units of the power spectrum) required for a peak to be recognized.
        This can be useful for filtering out noise or insignificant peaks, especially when a "knee" is present
        in the original data, which may persist in the periodic spectrum. Default is 0.01.
    peak_width_limits : tuple of (float, float), optional
        The lower and upper bounds for peak widths, in Hz. This helps in constraining the peak detection to
        meaningful features. Default is (0.5, 12.0).

    Returns
    -------
    pd.DataFrame
        A DataFrame containing the detected peak parameters for each channel. The DataFrame includes the
        following columns:
        - 'ch_name': Channel name
        - 'cf': Center frequency of the peak
        - 'bw': Bandwidth of the peak
        - 'pw': Peak height (power)


    Notes
    -----
    The function works by first optionally smoothing the periodic spectrum using a Savitzky-Golay filter.
    Then, it performs peak detection using the `scipy.signal.find_peaks` function, taking into account the
    specified peak thresholds and width limits. Peaks that do not meet the minimum height requirement are
    filtered out.

    The `cut_spectrum` parameter can be used to focus peak detection on a specific frequency range, which is
    particularly useful when the region of interest is known in advance.

    """

    if np.isnan(periodic_spectrum).sum() > 0:
        raise ValueError('peak width detection does not work properly with nans')

    freq_step = freqs[1] - freqs[0]

    # generate channel names if not given
    if ch_names is None:
        ch_names = [str(i) for i in np.arange(periodic_spectrum.shape[0])]

    # cut data
    if cut_spectrum is not None:
        freq_range = np.logical_and(freqs > cut_spectrum[0], freqs < cut_spectrum[1])
        freqs = freqs[freq_range]
        periodic_spectrum = periodic_spectrum[:, freq_range]

    # filter signal to get a smoother spectrum for peak extraction
    if smooth:
        window_length = int(smoothing_window // freq_step)
        assert window_length > polyorder, (
            'The smoothing window is too small you either need to increase \n'
            '`smoothing_window` or decrease the `polyorder`.'
        )

        filtered_spectrum = dsp.savgol_filter(periodic_spectrum, window_length=window_length, polyorder=polyorder)
    else:
        filtered_spectrum = periodic_spectrum
    # zero out negative values
    # filtered_spectrum[filtered_spectrum < 0] = 0
    # filtered_spectrum = np.log10(np.abs(filtered_spectrum.min()) + filtered_spectrum + 1e-1)
    # filtered_spectrum += np.abs(filtered_spectrum.min())
    # do peak finding on a channel by channel basis
    peak_list = []
    for ix, ch_name in enumerate(ch_names):
        peaks, peak_dict = dsp.find_peaks(
            filtered_spectrum[ix],
            height=[filtered_spectrum[ix].min(), filtered_spectrum[ix].max()],
            width=peak_width_limits / freq_step,  # in frequency in hz
            prominence=peak_threshold * np.std(filtered_spectrum[ix]),  # threshold in sd
            rel_height=0.75,  # relative peak height based on width
        )

        peak_list.append(
            pd.DataFrame(
                {
                    'ch_name': ch_name,
                    'cf': freqs[peaks],
                    'bw': peak_dict['widths'] * freq_step,
                    'pw': peak_dict['peak_heights'],
                }
            )
        )
    # combine & return
    if len(peak_list) >= 1:
        df_peaks = pd.concat(peak_list)
        # filter for peakheight
        df_peaks = df_peaks.query(f'pw > {min_peak_height}')
    else:
        df_peaks = pd.DataFrame({'ch_name': ch_names, 'cf': np.nan, 'bw': 0, 'pw': 0})
        # 0 is reasonable for power and bandwidth as "no" peaks are detected
        # therefore no power and bandwidth

    return df_peaks


# % find peaks in irasa sprint


def get_peak_params_sprint(
    periodic_spectrum: np.ndarray,
    freqs: np.ndarray,
    times: np.ndarray,
    ch_names: Iterable | None = None,
    smooth: bool = True,
    smoothing_window: int | float = 2,
    polyorder: int = 1,
    cut_spectrum: tuple[float, float] | None = None,
    peak_threshold: int | float = 1,
    min_peak_height: float = 0.01,
    peak_width_limits: tuple[float, float] = (0.5, 12),
) -> pd.DataFrame:
    """
    Extracts peak parameters from a periodic spectrogram obtained via IRASA.

    This function processes a time-resolved periodic spectrum to identify and extract peak parameters such as
    center frequency (cf), bandwidth (bw), and peak height (pw) for each time point. It applies smoothing,
    peak detection, and thresholding according to user-defined parameters
    (see get_peak_params for additional Information).

    Parameters
    ----------
    periodic_spectrum : np.ndarray
        2 or 3D array containing power values of the periodic spectrogram (shape: [Channels, Frequencies, Time Points]).
    freqs : np.ndarray
        1D array containing frequency values corresponding to the periodic spectrogram.
    times : np.ndarray
        1D array containing time points corresponding to the periodic spectrogram.
    ch_names : Iterable or None, optional
        List of channel names corresponding to the periodic spectrogram. If None, channels are labeled numerically.
        Default is None.
    smooth : bool, optional
        Whether to smooth the spectrum before peak extraction. Smoothing can help in reducing noise and better
        identifying peaks. Default is True.
    smoothing_window : int or float, optional
        The size of the smoothing window in Hz, passed to the Savitzky-Golay filter. Default is 2 Hz.
    polyorder : int, optional
        The polynomial order for the Savitzky-Golay filter used in smoothing. The polynomial order must be less
        than the window length. Default is 1.
    cut_spectrum : tuple of (float, float) or None, optional
        Tuple specifying the frequency range (lower_bound, upper_bound) to which the spectrum should be cut before
        peak extraction. If None, the full frequency range is used. Default is (1, 40).
    peak_threshold : int or float, optional
        Relative threshold for detecting peaks, defined as a multiple of the standard deviation of the filtered
        spectrum. Default is 1.
    min_peak_height : float, optional
        The minimum peak height (in absolute units of the power spectrum) required for a peak to be recognized.
        This can be useful for filtering out noise or insignificant peaks, especially when a "knee" is present
        in the data. Default is 0.01.
    peak_width_limits : tuple of (float, float), optional
        The lower and upper bounds for peak widths, in Hz. This helps in constraining the peak detection to
        meaningful features. Default is (0.5, 12).

    Returns
    -------
    pd.DataFrame
        A DataFrame containing the detected peak parameters for each channel and time point. The DataFrame
        includes the following columns:
        - 'ch_name': Channel name
        - 'cf': Center frequency of the peak
        - 'bw': Bandwidth of the peak
        - 'pw': Peak height (power)
        - 'time': Corresponding time point for the peak

    Notes
    -----
    This function iteratively processes each time point in the spectrogram, applying the `get_peak_params`
    function to extract peak parameters at each time point. The resulting peak parameters are combined into
    a single DataFrame.

    The function is particularly useful for analyzing time-varying spectral features, such as in dynamic or
    non-stationary M/EEG data, where peaks may shift in frequency, bandwidth, or amplitude over time.

    """

    time_list = []

    for ix, t in enumerate(times):
        cur_df = get_peak_params(
            periodic_spectrum[:, :, ix],
            freqs,
            ch_names=ch_names,
            smooth=smooth,
            smoothing_window=smoothing_window,
            polyorder=polyorder,
            cut_spectrum=cut_spectrum,
            peak_threshold=peak_threshold,
            min_peak_height=min_peak_height,
            peak_width_limits=peak_width_limits,
        )
        cur_df['time'] = t
        time_list.append(cur_df)

    df_time = pd.concat(time_list)

    return df_time


# %% find peaks irasa style
def get_band_info(df_peaks: pd.DataFrame, freq_range: tuple[int, int], ch_names: list) -> pd.DataFrame:
    """
    Extract peak information within a specified frequency range from a DataFrame of peak parameters.

    This function filters peaks found in the periodic spectrum or spectrogram to those within
    a specified frequency range.It ensures that every channel is represented in the output,
    filling in missing channels (i.e., channels without detected peaks in the specified range) with NaN values.

    Parameters
    ----------
    df_peaks : pd.DataFrame
        DataFrame containing peak parameters obtained from the `get_peak_params` function.
        The DataFrame should include columns for 'ch_name' (channel name), 'cf' (center frequency),
        'bw' (bandwidth), and 'pw' (peak height).
    freq_range : tuple of (int, int)
        Tuple specifying the lower and upper frequency bounds (in Hz) to filter peaks by. Only peaks
        with center frequencies (cf) within this range will be included in the output.
    ch_names : list
        List of channel names used in the computation of the periodic spectrum. This list ensures that
        every channel is accounted for in the output, even if no peaks were found in the specified range
        for certain channels.

    Returns
    -------
    pd.DataFrame
        DataFrame containing the peak parameters ('cf', 'bw', 'pw') for each channel within the specified
        frequency range. Channels without detected peaks in this range will have NaN values for these parameters.
        The DataFrame includes:
        - 'ch_name': Channel name
        - 'cf': Center frequency of the peak within the specified range
        - 'bw': Bandwidth of the peak within the specified range
        - 'pw': Peak height (power) within the specified range

    Notes
    -----
    This function is useful for isolating and analyzing peaks that occur within specific canonical frequency bands
    (e.g., alpha, beta, gamma) across multiple channels in a periodic spectrum. The inclusion of NaN
    entries for channels without detected peaks ensures that the output DataFrame is complete and aligned
    with the original channel list.

    """

    df_range = df_peaks.query(f'cf > {freq_range[0]}').query(f'cf < {freq_range[1]}')

    # we dont always get a peak in a queried range lets give those channels a nan
    missing_channels = list(set(ch_names).difference(df_range['ch_name'].unique()))
    missing_df = pd.DataFrame(np.nan, index=np.arange(len(missing_channels)), columns=['ch_name', 'cf', 'bw', 'pw'])
    missing_df['ch_name'] = missing_channels

    df_band_peaks = pd.concat([df_range, missing_df]).reset_index().drop(columns='index')

    return df_band_peaks



#%%
from neurodsp.sim import sim_combined
from neurodsp.utils import create_times
import scipy.signal as dsp
import matplotlib.pyplot as plt
import seaborn as sns

sns.set_style('ticks')
sns.set_context('poster')
# %%
fs = 500
n_seconds = 60
duration=2
overlap=0.5

sim_components = {'sim_powerlaw': {'exponent' : -1}, 
                  'sim_oscillation': {'freq' : 10}}
# %%

sig = sim_combined(n_seconds=n_seconds, fs=fs, components=sim_components)
times = create_times(n_seconds=n_seconds, fs=fs)
# %%
max_times = times < 1
f, axes = plt.subplots(ncols=2, figsize=(8, 4))
axes[0].plot(times[max_times], sig[max_times])
axes[0].set_ylabel('Amplitude (a.u.)')
axes[0].set_xlabel('Time (s)')
freq, psd = dsp.welch(sig, fs=fs, nperseg=duration*fs, noverlap=duration*fs*overlap)
axes[1].loglog(freq, psd)
axes[1].set_ylabel('Power (a.u.)')
axes[1].set_xlabel('Frequency (Hz)')

plt.tight_layout()


from pyrasa.irasa import irasa
# %%
freq_irasa, psd_ap, psd_p = irasa(sig, 
                                  fs=fs, 
                                  band=(1, 150), 
                                  kwargs_psd={'nperseg': duration*fs, 
                                             'noverlap': duration*fs*overlap
                                            },
                                hset_info=(1, 2, 0.05))
# %%
f, axes = plt.subplots(ncols=2, figsize=(8, 4))
axes[0].set_title('Periodic')
axes[0].plot(freq_irasa, psd_p[0,:])
axes[0].set_ylabel('Power (a.u.)')
axes[0].set_xlabel('Frequency (Hz)')
axes[1].set_title('Aperiodic')
axes[1].loglog(freq_irasa, psd_ap[0,:])
axes[1].set_ylabel('Power (a.u.)')
axes[1].set_xlabel('Frequency (Hz)')

f.tight_layout()
# %% get periodic stuff
from pyrasa.utils.peak_utils import get_peak_params
get_peak_params(psd_p, freqs=freq_irasa)
# %% get aperiodic stuff
from pyrasa.utils.aperiodic_utils import compute_slope
ap_params, gof_params = compute_slope(aperiodic_spectrum=psd_ap,
                                       freqs=freq_irasa, fit_func='knee')
ap_params
# %%
gof_params
# %% Lets check the knee
knee_freq = 15
knee = knee_freq ** (2*.5 + 1.5)

sim_components = {'sim_knee': {'exponent1' : -.5, 'exponent2': -1.5, 'knee':knee }, 
                  'sim_oscillation': {'freq' : 10}}

sig = sim_combined(n_seconds=n_seconds, fs=fs, components=sim_components)
# %%
max_times = times < 1
f, axes = plt.subplots(ncols=2, figsize=(8, 4))
axes[0].plot(times[max_times], sig[max_times])
axes[0].set_ylabel('Amplitude (a.u.)')
axes[0].set_xlabel('Time (s)')
freq, psd = dsp.welch(sig, fs=fs, nperseg=duration*fs, noverlap=duration*fs*overlap)
axes[1].loglog(freq, psd)
axes[1].set_ylabel('Power (a.u.)')
axes[1].set_xlabel('Frequency (Hz)')

plt.tight_layout()
# %%
freq_irasa, psd_ap, psd_p = irasa(sig, 
                                  fs=fs, 
                                  band=(1, 150), 
                                  kwargs_psd={'nperseg': duration*fs, 
                                             'noverlap': duration*fs*overlap
                                            },
                                hset_info=(1, 2, 0.025))
# %%
f, axes = plt.subplots(ncols=2, figsize=(8, 4))
axes[0].set_title('Periodic')
axes[0].plot(freq_irasa, psd_p[0,:])
axes[0].set_ylabel('Power (a.u.)')
axes[0].set_xlabel('Frequency (Hz)')
axes[1].set_title('Aperiodic')
axes[1].loglog(freq_irasa, psd_ap[0,:])
axes[1].set_ylabel('Power (a.u.)')
axes[1].set_xlabel('Frequency (Hz)')

f.tight_layout()

# %% get periodic stuff
from pyrasa.utils.peak_utils import get_peak_params
get_peak_params(psd_p, freqs=freq_irasa)
# %% get aperiodic stuff
from pyrasa.utils.aperiodic_utils import compute_slope
ap_params, gof_params = compute_slope(aperiodic_spectrum=psd_ap, freqs=freq_irasa, fit_func='fixed')
ap_params
# %%
gof_params
# %%
ap_params, gof_params = compute_slope(aperiodic_spectrum=psd_ap, freqs=freq_irasa, fit_func='knee')
ap_params
# %% 
gof_params
# %%



# %%
from neurodsp.sim import sim_powerlaw, sim_peak_oscillation, sim_combined, sim_knee, sim_bursty_oscillation

from cluster_jobs.meta_job import Job
import joblib
from neurodsp.utils.data import create_times
from neurodsp.sim import set_random_seed

from pyrasa.irasa import irasa, irasa_sprint
from pyrasa.utils.peak_utils import get_peak_params_sprint, get_peak_params
from pyrasa.utils.aperiodic_utils import compute_slope

import scipy.signal as dsp
import numpy as np
import pandas as pd


# %% Collect together parameters for combined signals


class simple_simulation(Job):
    job_data_folder = 'sim_all_simple11'
    exclude_kwargs_from_filename = ['n_secs', 'exp', 'knee', 'freq', 'bw', 'height', 'b_proba']
    include_hash_in_fname = True

    def run(
        self,
        subject_id,
        n_secs=60,
        fs=500,
        exp=-1.0,
        knee_freq=15,
        freq=10,
        bw=1.5,
        height=1.5,
        fit_type='fixed',
        b_proba=0.2,
        min_peak_height=0.01,
        burst_tol_time=0.2,
        n_seeds=30,
    ):
        kind = subject_id
        burst_list, p_ir_l, ap_ir_l = [], [], []
        p_sp_l_f, ap_sp_l_f = [], []
        for seed_val in np.arange(n_seeds):
            kwargs_psd = {'nperseg': fs * 4, 'noverlap': fs * 2}
            set_random_seed(seed_val=seed_val)
            burst_detection = None
            # specparam cannot yet work with double exponents setting exp_1 to 0 for now

            sim_feats = ['sim_exp', 'sim_knee_freq', 'sim_freq', 'sim_bw', 'sim_height', 'sim_b_proba', 'seed', 'kind']
            sim_params = dict(zip(sim_feats, np.zeros(len(sim_feats)) * np.nan))

            if kind == 'exponent_delta':
                sig = sim_powerlaw(n_seconds=n_secs, fs=fs, exponent=exp)
                sim_params['sim_exp'] = exp
                sim_params['kind'] = kind

            elif kind == 'knee_delta':
                exponent2 = -2
                sig = sim_knee(
                    n_seconds=n_secs, fs=fs, exponent1=0, exponent2=exponent2, knee=knee_freq ** np.abs(exponent2)
                )

                sim_params['sim_exp'] = exponent2
                sim_params['sim_knee_freq'] = knee_freq
                sim_params['kind'] = kind

            elif kind == 'exponent_knee_delta':
                knee_freq = 15
                sig = sim_knee(n_seconds=n_secs, fs=fs, exponent1=0, exponent2=exp, knee=knee_freq ** np.abs(exp))

                sim_params['sim_exp'] = exp
                sim_params['sim_knee_freq'] = knee_freq
                sim_params['kind'] = kind

            elif kind == 'burst_delta':
                # here we want to test irasa and irasa sprint
                # TODO: mark burst times.
                exp = -1.5
                freq = 10
                sig_ap = sim_powerlaw(n_seconds=n_secs, fs=fs, exponent=exp)
                sig_bursty = sim_bursty_oscillation(
                    n_seconds=n_secs, fs=fs, freq=freq, burst_params={'enter_burst': b_proba, 'leave_burst': b_proba}
                )
                burst_bool = sig_bursty > 0
                sig = sig_bursty + sig_ap
                sim_params['sim_exp'] = exp
                sim_params['sim_freq'] = freq
                sim_params['sim_b_proba'] = b_proba
                sim_params['kind'] = kind

                # burst detection using pyrasa
                _, periodic, freqs, times = irasa_sprint(
                    sig[np.newaxis, :], fs=fs, band=(1, 100), smooth=False, n_avgs=[3, 5, 7]
                )
                df_peaks = get_peak_params_sprint(periodic, freqs=freqs, times=times, min_peak_height=min_peak_height)

                burst_time = create_times(n_seconds=n_secs, fs=fs)
                burst_ts = burst_time[burst_bool]
                fair_bursts = burst_ts[(burst_ts / 0.02) % 1 == 0]  # need to have similar freq resolution

                detected = []
                for b in fair_bursts:
                    close_list = []
                    for p in df_peaks['time'].to_numpy():
                        close_list.append(np.isclose(p, b, atol=burst_tol_time))
                    detected.append(np.any(close_list))

                burst_detection = {
                    'percent_cycles_detected': np.mean(detected),
                    'n_cycles_detected': df_peaks['time'].to_numpy().shape[0],
                    'n_cycles_simulated': fair_bursts.shape[0],
                    'cycle_detection_ratio': fair_bursts.shape[0] / df_peaks['time'].to_numpy().shape[0],
                }

            elif kind == 'freq_width_delta':
                height = 1.5
                freq = 10
                exp = -1.5
                peak_params = {'freq': freq, 'bw': bw, 'height': height}

                sig_ap = sim_powerlaw(n_seconds=n_secs, fs=fs, exponent=exp)
                sig = sim_peak_oscillation(sig_ap, fs=fs, **peak_params)

                sim_params['sim_exp'] = exp
                sim_params['sim_freq'] = freq
                sim_params['sim_bw'] = bw
                sim_params['sim_height'] = height
                sim_params['kind'] = kind

            elif kind == 'freq_height_delta':
                bw = 0.5
                freq = 10
                exp = -1.5
                peak_params = {'freq': freq, 'bw': bw, 'height': height}

                sig_ap = sim_powerlaw(n_seconds=n_secs, fs=fs, exponent=exp)
                sig = sim_peak_oscillation(sig_ap, fs=fs, **peak_params)

                sim_params['sim_exp'] = exp
                sim_params['sim_freq'] = freq
                sim_params['sim_bw'] = bw
                sim_params['sim_height'] = height
                sim_params['kind'] = kind

            elif kind == 'freq_delta':
                comps = {
                    'sim_powerlaw': {
                        'exponent': -1.5,
                    },
                    'sim_oscillation': {'freq': freq},
                }
                sig = sim_combined(n_seconds=n_secs, fs=fs, components=comps)

                sim_params['sim_exp'] = exp
                sim_params['sim_freq'] = freq
                sim_params['sim_bw'] = bw
                sim_params['sim_height'] = height
                sim_params['kind'] = kind

            sim_params['seed'] = seed_val
            df_sim = pd.DataFrame(sim_params, index=[0])

            burst_list.append(burst_detection)

            # do irasa testing
            freqs, aperiodic, periodic = irasa(sig, fs=fs, band=(1, 100), kwargs_psd=kwargs_psd)

            df_ap_f, _ = compute_slope(aperiodic, freqs, scale=False, fit_func=fit_type)
            df_ap_f['method'] = 'pyrasa'
            df_ap_f = pd.concat([df_ap_f, df_sim], axis=1)

            df_p = get_peak_params(periodic, freqs, min_peak_height=min_peak_height, peak_width_limits=(1, 8))
            df_p['method'] = 'pyrasa'
            df_sim_p = pd.DataFrame(np.repeat(df_sim.values, df_p.shape[0], axis=0), columns=df_sim.columns)
            df_p = pd.concat([df_p, df_sim_p], axis=1)

            p_ir_l.append(df_p)
            ap_ir_l.append(df_ap_f)

            # do specparam testing
            from fooof import FOOOF

            fm = FOOOF(
                peak_width_limits=(1, 8),
                max_n_peaks=8,
                min_peak_height=0.1,
                aperiodic_mode=fit_type,
            )

            freqs, psd = dsp.welch(sig, fs=fs, **kwargs_psd)

            df_ap_sp_f, df_p_sp_f = self.extract_specparams(freqs, psd, fm, fit_type=fit_type, df_sim=df_sim)
            p_sp_l_f.append(df_p_sp_f)
            ap_sp_l_f.append(df_ap_sp_f)

            data = {
                'specparam': {
                    'periodic': pd.concat(p_sp_l_f),
                    'aperiodic': pd.concat(ap_sp_l_f),
                },
                'pyrasa': {
                    'periodic': pd.concat(p_ir_l),
                    'aperiodic': pd.concat(ap_ir_l),
                },
                'burst_pyrasa': burst_list,
            }

            joblib.dump(data, self.full_output_path)

    def extract_specparams(self, freqs, psd, fm, fit_type, df_sim):
        # freq_range adjusted to match irasas evaluated frequency range
        # i.e. 2*fmax, fmin/2
        fm.fit(freqs, psd, freq_range=(0.5, 200))
        # aperiodics
        if fit_type == 'fixed':
            df_ap = pd.DataFrame(
                {
                    'Offset': fm.get_params('aperiodic_params')[0],
                    'Exponent': fm.get_params('aperiodic_params')[1],
                },
                index=[0],
            )

        elif fit_type == 'knee':
            df_ap = pd.DataFrame(
                {
                    'Offset': fm.get_params('aperiodic_params')[0],
                    'Knee': fm.get_params('aperiodic_params')[1],
                    'Exponent_2': fm.get_params('aperiodic_params')[2],
                },
                index=[0],
            )

            df_ap['Knee Frequency (Hz)'] = df_ap['Knee'] ** (1.0 / df_ap['Exponent_2'])

        df_ap['fit_type'] = fit_type
        df_ap['method'] = 'specparam'
        # periodics

        peak_params = fm.get_params('peak_params')
        print(peak_params)
        if np.isscalar(peak_params[0]):
            df_p = pd.DataFrame(
                {
                    'cf': np.nan,
                    'pw': np.nan,
                    'bw': np.nan,
                },
                index=[0],
            )
        else:
            df_p = pd.DataFrame(
                peak_params,
                columns=('cf', 'pw', 'bw'),
            )

        df_p['method'] = 'specparam'
        df_sim_p = pd.DataFrame(np.repeat(df_sim.values, df_p.shape[0], axis=0), columns=df_sim.columns)

        df_p = pd.concat([df_p, df_sim_p], axis=1)
        df_ap = pd.concat([df_ap, df_sim], axis=1)

        return df_ap, df_p



import numpy as np
import pytest
from neurodsp.utils.sim import set_random_seed

from pyrasa.irasa import irasa_sprint
from pyrasa.utils.peak_utils import get_band_info

from .settings import EXPONENT, FS, MIN_R2_SPRINT, SPRINT_TOLERANCE

set_random_seed(42)


@pytest.mark.parametrize('fs', FS, scope='session')
@pytest.mark.parametrize('exponent_1', EXPONENT, scope='session')
@pytest.mark.parametrize('exponent_2', EXPONENT, scope='session')
def test_irasa_sprint(ts4sprint, fs, exponent_1, exponent_2):
    irasa_tf = irasa_sprint(
        ts4sprint[np.newaxis, :],
        win_duration=0.5,
        overlap_fraction=0.98,
        fs=fs,
        band=(0.1, 50),
        hset_info=(1.05, 4.0, 0.05),
    )

    # check basic aperiodic detection
    slope_fit = irasa_tf.fit_aperiodic_model(fit_func='fixed')
    #       irasa_tf.aperiodic[np.newaxis, :, :], freqs=irasa_tf.freqs, times=irasa_tf.time,
    #  )

    assert slope_fit.gof['R2'].mean() > MIN_R2_SPRINT
    assert np.isclose(
        np.mean(slope_fit.aperiodic_params.query('time < 7')['Exponent']), np.abs(exponent_1), atol=SPRINT_TOLERANCE
    )
    assert np.isclose(
        np.mean(slope_fit.aperiodic_params.query('time > 7')['Exponent']), np.abs(exponent_2), atol=SPRINT_TOLERANCE
    )

    # check basic peak detection
    df_peaks = irasa_tf.get_peaks(
        cut_spectrum=(1, 40),
        smooth=True,
        smoothing_window=3,
        min_peak_height=0.01,
        peak_width_limits=(0.5, 12),
    )

    df_alpha = get_band_info(df_peaks, freq_range=(8, 12), ch_names=[])
    alpha_peaks = df_alpha.query('pw > 0.10')
    t_diff = 0.025

    alpha_ts = alpha_peaks['time'].to_numpy()
    n_peaks = 0
    for ix, i in enumerate(alpha_ts):
        try:
            diff = alpha_ts[ix + 1] - i
            if diff > t_diff:
                n_peaks += 1
        except IndexError:
            pass

    # one missing burst is ok for now
    assert np.isclose(n_peaks, 7, atol=4)

    df_beta = get_band_info(df_peaks, freq_range=(20, 30), ch_names=[])
    beta_peaks = df_beta.query('pw > 0.10')

    beta_ts = beta_peaks['time'].to_numpy()
    n_peaks = 0
    for ix, i in enumerate(beta_ts):
        try:
            diff = beta_ts[ix + 1] - i
            if diff > t_diff:
                n_peaks += 1
        except IndexError:
            pass

    # one missing burst is ok for now
    assert np.isclose(n_peaks, 11, atol=4)


# test settings
@pytest.mark.parametrize('fs', [1000], scope='session')
@pytest.mark.parametrize('exponent_1', [-1], scope='session')
@pytest.mark.parametrize('exponent_2', [-2], scope='session')
def test_irasa_sprint_settings(ts4sprint, fs):
    # test dpss
    import scipy.signal as dsp

    irasa_sprint(
        ts4sprint[np.newaxis, :],
        fs=fs,
        band=(0.1, 100),
        win_func=dsp.windows.dpss,
    )

    # test too much bandwidht
    with pytest.raises(ValueError):
        irasa_sprint(
            ts4sprint[np.newaxis, :],
            fs=fs,
            band=(1, 100),
            win_func=dsp.windows.dpss,
            dpss_settings_time_bandwidth=4,
        )

    # test ratios
    with pytest.raises(ValueError):
        irasa_sprint(
            ts4sprint[np.newaxis, :],
            fs=fs,
            band=(1, 100),
            win_func=dsp.windows.dpss,
            dpss_settings_time_bandwidth=4,
        )



"""Classes for the MNE Python interface."""

# %% inherit from spectrum array

import matplotlib
import mne
import numpy as np
import pandas as pd
from attrs import define
from mne.time_frequency import EpochsSpectrumArray, SpectrumArray

from pyrasa.utils.aperiodic_utils import compute_aperiodic_model
from pyrasa.utils.peak_utils import get_peak_params
from pyrasa.utils.types import AperiodicFit


class PeriodicSpectrumArray(SpectrumArray):
    """Subclass of SpectrumArray"""

    def __init__(
        self: SpectrumArray,
        data: np.ndarray,
        info: mne.Info,
        freqs: np.ndarray,
        *,
        verbose: bool | str | int | None = None,
    ) -> None:
        # _check_data_shape(data, freqs, info, ndim=2)

        self.__setstate__(
            dict(
                method='IRASA',
                data=data,
                sfreq=info['sfreq'],
                dims=('channel', 'freq'),
                freqs=freqs,
                inst_type_str='Raw',
                data_type='Periodic Power Spectrum',
                info=info,
            )
        )

    def plot(
        self: SpectrumArray,
        *,
        picks: str | np.ndarray | slice | None = None,
        average: bool = False,
        dB: bool = False,  # noqa N803
        amplitude: bool | str = False,
        xscale: str = 'linear',
        ci: float | str | None = 'sd',
        ci_alpha: float = 0.3,
        color: str | tuple = 'black',
        alpha: float | None = None,
        spatial_colors: bool = True,
        sphere: float | np.ndarray | mne.bem.ConductorModel | None | str = None,
        exclude: tuple | list | str = (),
        axes: matplotlib.axes.Axes | list | None = None,
        show: bool = True,
    ) -> None:
        super().plot(
            picks=picks,
            average=average,
            dB=dB,
            amplitude=amplitude,
            xscale=xscale,
            ci=ci,
            ci_alpha=ci_alpha,
            color=color,
            alpha=alpha,
            spatial_colors=spatial_colors,
            sphere=sphere,
            exclude=exclude,
            axes=axes,
            show=show,
        )

    def plot_topo(
        self: SpectrumArray,
        *,
        dB: bool = False,  # noqa N803
        layout: mne.channels.Layout | None = None,
        color: str | tuple = 'w',
        fig_facecolor: str | tuple = 'k',
        axis_facecolor: str | tuple = 'k',
        axes: matplotlib.axes.Axes | list | None = None,
        block: bool = False,
        show: bool = True,
    ) -> None:
        super().plot_topo(
            dB=dB,
            layout=layout,
            color=color,
            fig_facecolor=fig_facecolor,
            axis_facecolor=axis_facecolor,
            axes=axes,
            block=block,
            show=show,
        )

    def get_peaks(
        self: SpectrumArray,
        smooth: bool = True,
        smoothing_window: float | int = 1,
        cut_spectrum: tuple[float, float] = (1, 40),
        peak_threshold: float = 2.5,
        min_peak_height: float = 0.0,
        polyorder: int = 1,
        peak_width_limits: tuple[float, float] = (0.5, 6),
    ) -> pd.DataFrame:
        """
        Extracts peak parameters from the periodic spectrum obtained via IRASA.

        This method identifies and extracts peak parameters such as center frequency (cf), bandwidth (bw),
        and peak height (pw) from a periodic spectrum using scipy's find_peaks function.
        The spectrum can be optionally smoothed prior to the peak detection.

        Parameters
        ----------
        smooth : bool, optional
            Whether to smooth the spectrum before peak extraction. Smoothing can help in reducing noise and
            better identifying peaks. Default is True.
        smoothing_window : int or float, optional
            The size of the smoothing window in Hz, passed to the Savitzky-Golay filter. Default is 1 Hz.
        polyorder : int, optional
            The polynomial order for the Savitzky-Golay filter used in smoothing. The polynomial order must be
            less than the window length. Default is 1.
        cut_spectrum : tuple of (float, float) or None, optional
            Tuple specifying the frequency range (lower_bound, upper_bound) to which the spectrum should be cut
            before peak extraction. If None, peaks are detected across the full frequency range. Default is None.
        peak_threshold : float, optional
            Relative threshold for detecting peaks, defined as a multiple of the standard deviation of the
            filtered spectrum. Default is 1.0.
        min_peak_height : float, optional
            The minimum peak height (in absolute units of the power spectrum) required for a peak to be recognized.
            This can be useful for filtering out noise or insignificant peaks, especially when a "knee" is present
            in the original data, which may persist in the periodic spectrum. Default is 0.01.
        peak_width_limits : tuple of (float, float), optional
            The lower and upper bounds for peak widths, in Hz. This helps in constraining the peak detection to
            meaningful features. Default is (0.5, 12.0).

        Returns
        -------
        pd.DataFrame
            A DataFrame containing the detected peak parameters for each channel. The DataFrame includes the
            following columns:
            - 'ch_name': Channel name
            - 'cf': Center frequency of the peak
            - 'bw': Bandwidth of the peak
            - 'pw': Peak height (power)


        Notes
        -----
        The function works by first optionally smoothing the periodic spectrum using a Savitzky-Golay filter.
        Then, it performs peak detection using the `scipy.signal.find_peaks` function, taking into account the
        specified peak thresholds and width limits. Peaks that do not meet the minimum height requirement are
        filtered out.

        The `cut_spectrum` parameter can be used to focus peak detection on a specific frequency range, which is
        particularly useful when the region of interest is known in advance.

        """

        peak_df = get_peak_params(
            self.get_data(),
            self.freqs,
            self.ch_names,
            smooth=smooth,
            smoothing_window=smoothing_window,
            cut_spectrum=cut_spectrum,
            peak_threshold=peak_threshold,
            min_peak_height=min_peak_height,
            polyorder=polyorder,
            peak_width_limits=peak_width_limits,
        )

        return peak_df


class AperiodicSpectrumArray(SpectrumArray):
    """Subclass of SpectrumArray"""

    def __init__(
        self: SpectrumArray,
        data: np.ndarray,
        info: mne.Info,
        freqs: np.ndarray,
        *,
        verbose: bool | str | int | None = None,
    ):
        # _check_data_shape(data, freqs, info, ndim=2)

        self.__setstate__(
            dict(
                method='IRASA',
                data=data,
                sfreq=info['sfreq'],
                dims=('channel', 'freq'),
                freqs=freqs,
                inst_type_str='Raw',
                data_type='Aperiodic Power Spectrum',
                info=info,
            )
        )

    def fit_aperiodic_model(
        self: SpectrumArray,
        fit_func: str = 'fixed',
        scale: bool = False,
        fit_bounds: tuple[float, float] | None = None,
    ) -> AperiodicFit:
        """
        Computes aperiodic parameters from the aperiodic spectrum using scipy's curve fitting function.

        This method can be used to model the aperiodic (1/f-like) component of the power spectrum. Per default,
        users can choose between a fixed or knee model fit or specify their own fit method see examples
        custom_fit_functions.ipynb for an example.
        The method returns the fitted parameters for each channel along with some goodness of fit metrics.

        Parameters
        ----------
        fit_func : str or type[AbstractFitFun], optional
            The fitting function to use. Can be "fixed" for a linear fit or "knee" for a fit that includes a
            knee (bend) in the spectrum or a class that is inherited from AbstractFitFun. The default is 'fixed'.
        ch_names : Iterable or None, optional
            Channel names corresponding to the aperiodic spectrum. If None, channels will be named numerically
            in ascending order. Default is None.
        scale : bool, optional
            Whether to scale the data to improve fitting accuracy. This is useful in cases where
            power values are very small (e.g., 1e-28), which may lead to numerical precision issues during fitting.
            After fitting, the parameters are rescaled to match the original data scale. Default is False.
        fit_bounds : tuple[float, float] or None, optional
            Tuple specifying the lower and upper frequency bounds for the fit function. If None, the entire frequency
            range is used. Otherwise, the spectrum is cropped to the specified bounds. Default is None.

        Returns
        -------
        AperiodicFit
            An object containing three pandas DataFrames:
                - aperiodic_params : pd.DataFrame
                    A DataFrame containing the fitted aperiodic parameters for each channel.
                - gof : pd.DataFrame
                    A DataFrame containing the goodness of fit metrics for each channel.
                - model : pd.DataFrame
                    A DataFrame containing the predicted aperiodic model for each channel and each time point.


        Notes
        -----
        This function fits the aperiodic component of the power spectrum using scipy's curve fitting function.
        The fitting can be performed using either a simple linear model ('fixed') or a more complex model
        that includes a "knee" point, where the spectrum bends. The resulting parameters can help in
        understanding the underlying characteristics of the aperiodic component in the data.

        If the `fit_bounds` parameter is used, it ensures that only the specified frequency range is considered
        for fitting, which can be important to avoid fitting artifacts outside the region of interest.

        The `scale` parameter can be crucial when dealing with data that have extremely small values,
        as it helps to mitigate issues related to machine precision during the fitting process.

        The function asserts that the input data are of the correct type and shape, and raises warnings
        if the first frequency value is zero, as this can cause issues during model fitting.
        """

        return compute_aperiodic_model(
            self.get_data(),
            self.freqs,
            ch_names=self.ch_names,
            scale=scale,
            fit_func=fit_func,
            fit_bounds=fit_bounds,
        )


# %%
class PeriodicEpochsSpectrum(EpochsSpectrumArray):
    """Subclass of EpochsSpectrumArray"""

    def __init__(
        self: EpochsSpectrumArray,
        data: np.ndarray,
        info: mne.Info,
        freqs: np.ndarray,
        events: np.ndarray | None = None,
        event_id: int | list | dict | str | None = None,
        *,
        verbose: bool | str | int | None = None,
    ) -> None:
        if events is not None and data.shape[0] != events.shape[0]:
            raise ValueError(
                f'The first dimension of `data` ({data.shape[0]}) must match the '
                f'first dimension of `events` ({events.shape[0]}).'
            )

        self.__setstate__(
            dict(
                method='IRASA',
                data=data,
                sfreq=info['sfreq'],
                dims=('epoch', 'channel', 'freq'),
                freqs=freqs,
                inst_type_str='Epochs',
                data_type='Periodic Power Spectrum',
                info=info,
                events=events,
                event_id=event_id,
                metadata=None,
                selection=np.arange(data.shape[0]),
                drop_log=tuple(tuple() for _ in range(data.shape[0])),
            )
        )

    def plot(
        self: EpochsSpectrumArray,
        *,
        picks: str | np.ndarray | slice | None = None,
        average: bool = False,
        dB: bool = False,  # noqa N803
        amplitude: bool = False,
        xscale: str = 'linear',
        ci: float | str | None = 'sd',
        ci_alpha: float = 0.3,
        color: str | tuple = 'black',
        alpha: float | None = None,
        spatial_colors: bool = True,
        sphere: float | np.ndarray | mne.bem.ConductorModel | None | str = None,
        exclude: list | tuple | str = (),
        axes: matplotlib.axes.Axes | list | None = None,
        show: bool = True,
    ) -> None:
        super().plot(
            picks=picks,
            average=average,
            dB=dB,
            amplitude=amplitude,
            xscale=xscale,
            ci=ci,
            ci_alpha=ci_alpha,
            color=color,
            alpha=alpha,
            spatial_colors=spatial_colors,
            sphere=sphere,
            exclude=exclude,
            axes=axes,
            show=show,
        )

    def plot_topo(
        self: EpochsSpectrumArray,
        *,
        dB: bool = False,  # noqa N803
        layout: mne.channels.Layout | None = None,
        color: str | tuple = 'w',
        fig_facecolor: str | tuple = 'k',
        axis_facecolor: str | tuple = 'k',
        axes: matplotlib.axes.Axes | list | None = None,
        block: bool = False,
        show: bool = True,
    ) -> None:
        super().plot_topo(
            dB=dB,
            layout=layout,
            color=color,
            fig_facecolor=fig_facecolor,
            axis_facecolor=axis_facecolor,
            axes=axes,
            block=block,
            show=show,
        )

    def get_peaks(
        self: EpochsSpectrumArray,
        smooth: bool = True,
        smoothing_window: float | int = 1,
        cut_spectrum: tuple[float, float] = (1.0, 40.0),
        peak_threshold: float = 2.5,
        min_peak_height: float = 0.0,
        polyorder: int = 1,
        peak_width_limits: tuple[float, float] = (0.5, 6.0),
    ) -> pd.DataFrame:
        """
        Extracts peak parameters from the periodic spectrum obtained via IRASA.

        This method identifies and extracts peak parameters such as center frequency (cf), bandwidth (bw),
        and peak height (pw) from a periodic spectrum using scipy's find_peaks function.
        The spectrum can be optionally smoothed prior to the peak detection.

        Parameters
        ----------
        smooth : bool, optional
            Whether to smooth the spectrum before peak extraction. Smoothing can help in reducing noise and
            better identifying peaks. Default is True.
        smoothing_window : int or float, optional
            The size of the smoothing window in Hz, passed to the Savitzky-Golay filter. Default is 1 Hz.
        polyorder : int, optional
            The polynomial order for the Savitzky-Golay filter used in smoothing. The polynomial order must be
            less than the window length. Default is 1.
        cut_spectrum : tuple of (float, float) or None, optional
            Tuple specifying the frequency range (lower_bound, upper_bound) to which the spectrum should be cut
            before peak extraction. If None, peaks are detected across the full frequency range. Default is None.
        peak_threshold : float, optional
            Relative threshold for detecting peaks, defined as a multiple of the standard deviation of the
            filtered spectrum. Default is 1.0.
        min_peak_height : float, optional
            The minimum peak height (in absolute units of the power spectrum) required for a peak to be recognized.
            This can be useful for filtering out noise or insignificant peaks, especially when a "knee" is present
            in the original data, which may persist in the periodic spectrum. Default is 0.01.
        peak_width_limits : tuple of (float, float), optional
            The lower and upper bounds for peak widths, in Hz. This helps in constraining the peak detection to
            meaningful features. Default is (0.5, 12.0).

        Returns
        -------
        pd.DataFrame
            A DataFrame containing the detected peak parameters for each channel. The DataFrame includes the
            following columns:
            - 'ch_name': Channel name
            - 'cf': Center frequency of the peak
            - 'bw': Bandwidth of the peak
            - 'pw': Peak height (power)


        Notes
        -----
        The function works by first optionally smoothing the periodic spectrum using a Savitzky-Golay filter.
        Then, it performs peak detection using the `scipy.signal.find_peaks` function, taking into account the
        specified peak thresholds and width limits. Peaks that do not meet the minimum height requirement are
        filtered out.

        The `cut_spectrum` parameter can be used to focus peak detection on a specific frequency range, which is
        particularly useful when the region of interest is known in advance.

        """

        event_dict = {val: key for key, val in self.event_id.items()}
        events = self.events[:, 2]

        peak_list = []
        for ix, cur_epoch in enumerate(self.get_data()):
            peak_df = get_peak_params(
                cur_epoch,
                self.freqs,
                self.ch_names,
                smooth=smooth,
                smoothing_window=smoothing_window,
                cut_spectrum=cut_spectrum,
                peak_threshold=peak_threshold,
                min_peak_height=min_peak_height,
                polyorder=polyorder,
                peak_width_limits=peak_width_limits,
            )

            peak_df['event_id'] = event_dict[events[ix]]

            peak_list.append(peak_df)

        return pd.concat(peak_list)


class AperiodicEpochsSpectrum(EpochsSpectrumArray):
    """Subclass of EpochsSpectrumArray"""

    def __init__(
        self: EpochsSpectrumArray,
        data: np.ndarray,
        info: mne.Info,
        freqs: np.ndarray,
        events: np.ndarray | None = None,
        event_id: int | list | dict | str | None = None,
        *,
        verbose: bool | str | int | None = None,
    ):
        if events is not None and data.shape[0] != events.shape[0]:
            raise ValueError(
                f'The first dimension of `data` ({data.shape[0]}) must match the '
                f'first dimension of `events` ({events.shape[0]}).'
            )

        self.__setstate__(
            dict(
                method='IRASA',
                data=data,
                sfreq=info['sfreq'],
                dims=('epoch', 'channel', 'freq'),
                freqs=freqs,
                inst_type_str='Epochs',
                data_type='Aperiodic Power Spectrum',
                info=info,
                events=events,
                event_id=event_id,
                metadata=None,
                selection=np.arange(data.shape[0]),
                drop_log=tuple(tuple() for _ in range(data.shape[0])),
            )
        )

    def fit_aperiodic_model(
        self: SpectrumArray,
        fit_func: str = 'fixed',
        scale: bool = False,
        fit_bounds: tuple[float, float] | None = None,
    ) -> AperiodicFit:
        """
        Computes aperiodic parameters from the aperiodic spectrum using scipy's curve fitting function.

        This method can be used to model the aperiodic (1/f-like) component of the power spectrum. Per default,
        users can choose between a fixed or knee model fit or specify their own fit method see examples
        custom_fit_functions.ipynb for an example.
        The method returns the fitted parameters for each channel along with some goodness of fit metrics.

        Parameters
        ----------
        fit_func : str or type[AbstractFitFun], optional
            The fitting function to use. Can be "fixed" for a linear fit or "knee" for a fit that includes a
            knee (bend) in the spectrum or a class that is inherited from AbstractFitFun. The default is 'fixed'.
        ch_names : Iterable or None, optional
            Channel names corresponding to the aperiodic spectrum. If None, channels will be named numerically
            in ascending order. Default is None.
        scale : bool, optional
            Whether to scale the data to improve fitting accuracy. This is useful in cases where
            power values are very small (e.g., 1e-28), which may lead to numerical precision issues during fitting.
            After fitting, the parameters are rescaled to match the original data scale. Default is False.
        fit_bounds : tuple[float, float] or None, optional
            Tuple specifying the lower and upper frequency bounds for the fit function. If None, the entire frequency
            range is used. Otherwise, the spectrum is cropped to the specified bounds. Default is None.

        Returns
        -------
        AperiodicFit
            An object containing three pandas DataFrames:
                - aperiodic_params : pd.DataFrame
                    A DataFrame containing the fitted aperiodic parameters for each channel.
                - gof : pd.DataFrame
                    A DataFrame containing the goodness of fit metrics for each channel.
                - model : pd.DataFrame
                    A DataFrame containing the predicted aperiodic model for each channel and each time point.


        Notes
        -----
        This function fits the aperiodic component of the power spectrum using scipy's curve fitting function.
        The fitting can be performed using either a simple linear model ('fixed') or a more complex model
        that includes a "knee" point, where the spectrum bends. The resulting parameters can help in
        understanding the underlying characteristics of the aperiodic component in the data.

        If the `fit_bounds` parameter is used, it ensures that only the specified frequency range is considered
        for fitting, which can be important to avoid fitting artifacts outside the region of interest.

        The `scale` parameter can be crucial when dealing with data that have extremely small values,
        as it helps to mitigate issues related to machine precision during the fitting process.

        The function asserts that the input data are of the correct type and shape, and raises warnings
        if the first frequency value is zero, as this can cause issues during model fitting.
        """

        event_dict = {val: key for key, val in self.event_id.items()}
        events = self.events[:, 2]

        aps_list, gof_list, pred_list = [], [], []
        for ix, cur_epoch in enumerate(self.get_data()):
            slope_fit = compute_aperiodic_model(
                cur_epoch,
                self.freqs,
                ch_names=self.ch_names,
                scale=scale,
                fit_func=fit_func,
                fit_bounds=fit_bounds,
            )

            slope_fit.aperiodic_params['event_id'] = event_dict[events[ix]]
            slope_fit.gof['event_id'] = event_dict[events[ix]]
            aps_list.append(slope_fit.aperiodic_params.copy())
            gof_list.append(slope_fit.gof.copy())
            pred_list.append(slope_fit.model)

        return AperiodicFit(aperiodic_params=pd.concat(aps_list), gof=pd.concat(gof_list), model=pd.concat(pred_list))


@define
class IrasaRaw:
    periodic: PeriodicSpectrumArray
    aperiodic: AperiodicSpectrumArray


@define
class IrasaEpoched:
    periodic: PeriodicEpochsSpectrum
    aperiodic: AperiodicEpochsSpectrum



#%%
from neurodsp.sim import sim_combined, sim_knee
import numpy as np
from os.path import join

n_secs=60
fs_list=[500, 750, 1000]
exp1=0
exp2_list=[-1., -1.5, -2.0]
knee_freq_list=[10, 15, 25]
osc_list = [5, 10, 20]

base_folder = '/Users/fabian.schmidt/git/pyrasa/tests/test_data/'

for fs in fs_list:
    for exp2 in exp2_list:
        for osc_freq in osc_list:
            for knee_freq in knee_freq_list:
                #% generate and save knee osc
                knee = knee_freq ** np.abs(exp2)
                components = {'sim_knee': {'exponent1': exp1, 'exponent2': exp2, 'knee': knee}, 
                            'sim_oscillation': {'freq': osc_freq}
                            }
                cmb_sim = sim_combined(n_seconds=n_secs, fs=fs, components=components)

                fname = f'cmb_sim__fs_{fs}__exp1_{np.abs(exp1)}__exp2_{np.abs(exp2)}_knee_{np.round(knee, 0)}__osc_freq_{osc_freq}_.npy'
                np.save(join(base_folder + 'knee_osc_data', fname), cmb_sim, allow_pickle=False)

                #% generate and save knee
                knee_sim = sim_knee(n_seconds=n_secs, fs=fs, exponent1=exp1, exponent2=exp2, knee=knee)

                fname = f'knee_sim__fs_{fs}__exp1_{np.abs(exp1)}__exp2_{np.abs(exp2)}_knee_{np.round(knee, 0)}_.npy'
                np.save(join(base_folder + 'knee_data', fname), knee_sim, allow_pickle=False)

# %%



# %%
from neurodsp.sim import sim_synaptic_current
from neurodsp.utils import create_times
import pandas as pd

import scipy.signal as dsp
import numpy as np

import matplotlib.pyplot as plt
import seaborn as sns

sns.set_style('ticks')
sns.set_context('poster')


# %% based on info from gao 2017)
def make_gao_ei(n_secs, fs, ei_ratio=4):
    """Do lfp gao style parameters as in Gao et al. 2017"""

    mv_rest = -0.65
    mv_rev_e, mv_rev_i = 0, -0.80
    ampa = sim_synaptic_current(n_secs, fs=fs, n_neurons=8000, firing_rate=2, tau_r=0.0001, tau_d=0.002, t_ker=1) * (
        mv_rest - mv_rev_e
    )
    gaba = sim_synaptic_current(n_secs, fs=fs, n_neurons=2000, firing_rate=5, tau_r=0.0005, tau_d=0.01, t_ker=1) * (
        mv_rest - mv_rev_i
    )

    lfp = ampa * ei_ratio + gaba

    return lfp


nsecs = 30
fs = 1000
times = create_times(nsecs, fs)
# %%
ratios = 1 / np.arange(1, 6)
lfps = [make_gao_ei(n_secs=nsecs, fs=fs, ei_ratio=i) for i in ratios]
freqs, psds = zip(*([dsp.welch(lfp, fs=fs, nperseg=4 * fs, noverlap=2 * fs) for lfp in lfps]))

# %%
cmap = sns.color_palette('Greys', np.shape(lfps)[0])
f, ax = plt.subplots(figsize=(4, 4))
for ix, psd in enumerate(psds):
    ax.loglog(freqs[0], psd, color=cmap[ix])

# %%
from fooof import FOOOFGroup

fg = FOOOFGroup(max_n_peaks=0, aperiodic_mode='knee')
fg.fit(freqs[0], np.array(psds))

df = pd.DataFrame(fg.get_params('aperiodic_params'), columns=['offset', 'knee', 'exponent'])

df['Knee Frequency'] = df['knee'] ** (1.0 / df['exponent'])
df['tau'] = 1.0 / (2 * np.pi * df['Knee Frequency'])
# df['ei'] = ['1/2', '1/3', '1/4', '1/5', '1/6']
df['ei'] = 1 / ratios
# %%
plt.scatter(df['ei'], df['Knee Frequency'])
# %%
df.corr('spearman')
# %%
df
# %%



# %%
import sys
from neurodsp.sim import set_random_seed
from neurodsp.sim import sim_powerlaw, sim_oscillation
from neurodsp.utils import create_times
from neurodsp.plts import plot_timefrequency  #

from neurodsp.timefrequency import compute_wavelet_transform
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import pandas as pd

import matplotlib as mpl

new_rc_params = {'text.usetex': False, 'svg.fonttype': 'none'}
mpl.rcParams.update(new_rc_params)


sns.set_style('ticks')
sns.set_context('talk')

set_random_seed(42)

from pyrasa.irasa import irasa_sprint

# %%
# Set some general settings, to be used across all simulations
fs = 500
n_seconds = 15
duration = 4
overlap = 0.5

# Create a times vector for the simulations
times = create_times(n_seconds, fs)

# %%
alpha = sim_oscillation(n_seconds=0.5, fs=fs, freq=10)
no_alpha = np.zeros(len(alpha))
beta = sim_oscillation(n_seconds=0.5, fs=fs, freq=25)
no_beta = np.zeros(len(beta))

exp_1 = sim_powerlaw(n_seconds=2.5, fs=fs, exponent=-1)
exp_2 = sim_powerlaw(n_seconds=2.5, fs=fs, exponent=-2)

# %%
alphas = np.concatenate([no_alpha, alpha, no_alpha, alpha, no_alpha])
betas = np.concatenate([beta, no_beta, beta, no_beta, beta])

sim_ts = np.concatenate(
    [
        exp_1 + alphas,
        exp_1 + alphas + betas,
        exp_1 + betas,
        exp_2 + alphas,
        exp_2 + alphas + betas,
        exp_2 + betas,
    ]
)
# %%
plt.plot(times, sim_ts)
# %%
freqs = np.arange(1, 100, 0.5)

# %%
mwt = compute_wavelet_transform(
    sim_ts,
    fs=fs,
    freqs=freqs,
    n_cycles=11,
)


# %%


# f.savefig('../results/time_freq_sim_mwt.svg')
# %%
sgramm_ap, sgramm_p, freqs_ir, times_ir = irasa_sprint(
    sim_ts[np.newaxis, :], fs=fs, band=(1, 100), freq_res=0.5, smooth=False, n_avgs=[3, 7, 11]
)

# %%
f, axes = plt.subplots(figsize=(14, 4), ncols=3)
mwt = np.abs(mwt)
plot_timefrequency(times, freqs, mwt, vmin=0, ax=axes[0])
plot_timefrequency(times_ir, freqs_ir, np.squeeze(sgramm_ap), vmin=0, ax=axes[1])
plot_timefrequency(times_ir, freqs_ir, np.squeeze(sgramm_p), vmin=0, ax=axes[2])

# f.savefig('../results/time_freq_split.svg')
# %% now extract the aperiodic features
from pyrasa.utils.aperiodic_utils import compute_slope_sprint

df_aps, df_gof = compute_slope_sprint(sgramm_ap[np.newaxis, :, :], freqs=freqs_ir, times=times_ir, fit_func='fixed')

# %%
ave = df_gof['r_squared'].mean()
std = df_gof['r_squared'].std()
print(f'Average goodness of fit {ave} and deviation is {std}')


# %%
ave = df_gof.query('time < 7')['r_squared'].mean()
std = df_gof.query('time < 7')['r_squared'].std()
print(f'Average goodness of fit {ave} and deviation is {std}')


# %%
ave = df_gof.query('time > 7')['r_squared'].mean()
std = df_gof.query('time > 7')['r_squared'].std()
print(f'Average goodness of fit {ave} and deviation is {std}')


# %%
f, ax = plt.subplots(nrows=3, figsize=(8, 7))

ax[0].plot(df_aps['time'], df_aps['Offset'])
ax[0].set_ylabel('Offset')
ax[0].set_xlabel('time (s)')
ax[1].plot(df_aps['time'], df_aps['Exponent'])
ax[1].set_ylabel('Exponent')
ax[1].set_xlabel('time (s)')
ax[2].plot(df_aps['time'], df_gof['r_squared'])
ax[2].set_ylabel('R2')
ax[2].set_xlabel('time (s)')

f.tight_layout()
# f.savefig('../results/time_resolved_exp_r2.svg')
# %%
# %%
from pyrasa.utils.peak_utils import get_peak_params_sprint

df_peaks = get_peak_params_sprint(
    sgramm_p[np.newaxis, :, :], freqs=freqs_ir, times=times_ir, smooth=True, min_peak_height=0.1
)

plot_timefrequency(times_ir, freqs_ir, sgramm_p, vmin=0)

# %% Plot peak results
f, ax = plt.subplots(nrows=3, figsize=(8, 7))

for ix, cur_key in enumerate(['cf', 'pw', 'bw']):
    ax[ix].plot(df_peaks['time'], df_peaks[cur_key])
    ax[ix].set_ylabel(cur_key)
    ax[ix].set_xlabel('time (s)')
    ax[ix].set_xlim(0, 15)

f.tight_layout()
# f.savefig('../results/time_resolved_peak_params.svg')
# %%
sys.path.append('/home/schmidtfa/git/SPRiNT')
from SPRiNT_py import SPRiNT_stft_py
# %%


opt = {
    'sfreq': fs,  # Input sampling rate
    'WinLength': 0.5,  # STFT window length
    'WinOverlap': 50,  # Overlap between sliding windows (in %)
    'WinAverage': 3,  # Number of overlapping windows being averaged
    'rmoutliers': 1,  # Apply peak post-processing
    'maxTime': 6,  # Maximum distance of nearby peaks in time (in n windows)
    'maxFreq': 2.5,  # Maximum distance of nearby peaks in frequency (in Hz)
    'minNear': 3,  # Minimum number of similar peaks nearby (using above bounds)
}

output = SPRiNT_stft_py(sim_ts, opt)

from fooof import FOOOFGroup, fit_fooof_3d


fg = FOOOFGroup(peak_width_limits=[1, 8], min_peak_height=0.01, max_n_peaks=8)
fgs = fit_fooof_3d(fg, output['freqs'], output['TF'])


df_fooof = pd.DataFrame(fgs[0].get_params('peak_params'), columns=('CF', 'PW', 'BW', 'Time'))
dict_a = dict(zip(df_fooof['Time'].unique().astype(int), output['ts']))
df_fooof['Time'] = df_fooof['Time'].replace(dict_a)
df_fooof['Time'] = df_fooof['Time'] / 2

df_alpha_fooof = df_fooof.query('CF < 12').query('CF > 8')
df_beta_fooof = df_fooof.query('CF < 30').query('CF > 20')
# %%
df_alpha_fooof
# %%
from pyrasa.utils.peak_utils import get_band_info

df_alpha = get_band_info(df_peaks, freq_range=(8, 12), ch_names=[])
df_beta = get_band_info(df_peaks, freq_range=(20, 30), ch_names=[])

# %%
f, ax = plt.subplots(figsize=(12, 4), ncols=2)

ax[0].plot(df_alpha['time'], df_alpha['pw'])
ax[1].plot(df_beta['time'], df_beta['pw'])

yax = ['Alpha Power (8-12Hz)', 'Beta Power (20-30Hz)']
for ix, c_ax in enumerate(ax):
    c_ax.set_xlabel('Time (s)')
    c_ax.set_ylabel(yax[ix])

sns.despine()
f.tight_layout()
# f.savefig('../results/detected_peaks_pyrasa.svg')

# %%
plt.plot(df_beta['time'], df_beta['pw'])


# %%

f, ax = plt.subplots(figsize=(12, 4), ncols=2)

ax[0].plot(df_alpha_fooof['Time'], df_alpha_fooof['PW'])
ax[1].plot(df_beta_fooof['Time'], df_beta_fooof['PW'])

yax = ['Alpha Power (8-12Hz)', 'Beta Power (20-30Hz)']
for ix, c_ax in enumerate(ax):
    c_ax.set_xlabel('Time (s)')
    c_ax.set_ylabel(yax[ix])

sns.despine()
f.tight_layout()
# f.savefig('../results/detected_peaks_sprint.svg')


# %%
df_alpha_fooof
# %%

# %%



# %%
import joblib
from pathlib import Path
from pyrasa.utils.peak_utils import get_peak_params, get_band_info
from pyrasa.utils.aperiodic_utils import compute_slope
import os
from pymatreader import read_mat
import scipy.signal as dsp

from pyrasa.irasa import irasa

from cluster_jobs.meta_job import Job

# %% load data


class eo_ec_analysis(Job):
    job_data_folder = 'eo_ec'

    def run(
        self,
        subject_id,
        condition,
    ):
        subject_id = int(subject_id)
        base_p = '/'.join(os.path.abspath(os.path.dirname(__file__)).split('/')[:-1])
        if subject_id < 10:
            cur_file = list(
                Path(base_p + '/data/SPIS-Resting-State-Dataset-master/Pre-SART EEG').glob(
                    f'S0{subject_id}*_{condition}.mat'
                )
            )[0]
        else:
            cur_file = list(
                Path(base_p + '/data/SPIS-Resting-State-Dataset-master/Pre-SART EEG').glob(
                    f'S{subject_id}*_{condition}.mat'
                )
            )[0]

        sig = read_mat(cur_file)['dataRest']
        fs = 256

        psd_kwargs = {'nperseg': fs * 4, 'noverlap': fs * 2, 'average': 'median'}

        ch_names = [
            'Fp1',
            'AF7',
            'AF3',
            'F1',
            'F3',
            'F5',
            'F7',
            'FT7',
            'FC5',
            'FC3',
            'FC1',
            'C1',
            'C3',
            'C5',
            'T7',
            'TP7',
            'CP5',
            'CP3',
            'CP1',
            'P1',
            'P3',
            'P5',
            'P7',
            'P9',
            'PO7',
            'PO3',
            'O1',
            'Iz',
            'Oz',
            'POz',
            'Pz',
            'CPz',
            'Fpz',
            'Fp2',
            'AF8',
            'AF4',
            'Afz',
            'Fz',
            'F2',
            'F4',
            'F6',
            'F8',
            'FT8',
            'FC6',
            'FC4',
            'FC2',
            'FCz',
            'Cz',
            'C2',
            'C4',
            'C6',
            'T8',
            'TP8',
            'CP6',
            'CP4',
            'CP2',
            'P2',
            'P4',
            'P6',
            'P8',
            'P10',
            'PO8',
            'PO4',
            'O26',
        ]

        # % compute raw spectrum
        freq, psd = dsp.welch(sig[:64, :], fs=fs, **psd_kwargs)

        # % compute irasa
        freq_irasa, psd_ap, psd_p = irasa(sig, fs=fs, band=(0.1, 30), kwargs_psd=psd_kwargs, hset_info=(1, 2, 0.05))
        # %
        peak_params = get_peak_params(psd_p, freqs=freq_irasa, ch_names=ch_names)
        alpha_df = get_band_info(peak_params, freq_range=(8, 12), ch_names=ch_names)
        alpha_df['s_id'] = subject_id
        alpha_df['condition'] = condition
        # %
        fixed, gof_f = compute_slope(psd_ap, freq_irasa, fit_func='fixed', scale=False, ch_names=ch_names)
        fixed['s_id'] = subject_id
        fixed['condition'] = condition
        gof_f['s_id'] = subject_id
        gof_f['condition'] = condition
        try:
            knee, gof_k = compute_slope(psd_ap, freq_irasa, fit_func='knee', scale=False, ch_names=ch_names)
            knee['s_id'] = subject_id
            knee['condition'] = condition
            gof_k['s_id'] = subject_id
            gof_k['condition'] = condition
        except ValueError:
            knee = None

        data = {
            'periodics': {'peak_params': peak_params, 'alpha': alpha_df},
            'aperiodics': {'fixed': fixed, 'knee': knee},
            'psd': {'psd': psd, 'freq': freq},
        }

        joblib.dump(data, self.full_output_path)



# %%
import joblib
from pathlib import Path
import os
import pandas as pd
import numpy as np

import matplotlib.pyplot as plt
import seaborn as sns

import matplotlib as mpl

new_rc_params = {'text.usetex': False, 'svg.fonttype': 'none'}
mpl.rcParams.update(new_rc_params)


sns.set_style('ticks')
sns.set_context('poster')


base_p = '/'.join(os.path.abspath(os.path.dirname(__file__)).split('/')[:-1])

all_files = list(Path(base_p + '/data/sim_all_f/broad_fixed').glob('*.dat'))
# %%
reload = True
if reload:
    df_ap_sim = pd.read_csv('../results/sim_broad_freq_fixed_ap.csv')
    df_p_sim = pd.read_csv('../results/sim_broad_freq_fixed_p.csv')

else:
    ap_list, p_list = [], []

    for f in all_files:
        cur_f = joblib.load(f)

        cur_p = pd.concat(
            [cur_f['specparam']['fixed']['periodic'], cur_f['pyrasa']['fixed']['periodic']]
        )  # .dropna(axis=1)

        cur_ap = pd.concat(
            [cur_f['specparam']['fixed']['aperiodic'], cur_f['pyrasa']['fixed']['aperiodic']]
        )  # .dropna(axis=1)

        ap_list.append(cur_ap)
        p_list.append(cur_p)

    df_ap_sim = pd.concat(ap_list).reset_index()
    df_ap_sim['sim_exp'] = df_ap_sim['sim_exp'].abs()
    df_ap_sim.to_csv('../results/sim_broad_freq_fixed_ap.csv')

    df_p_sim = pd.concat(p_list)
    df_p_sim['sim_exp'] = df_p_sim['sim_exp'].abs()

    df_p_sim.to_csv('../results/sim_broad_freq_fixed_p.csv')

# %%

df_ap_sim['sim_exp'] = df_ap_sim['sim_exp'].abs()
df_ap_sim['Exponent Δ'] = np.abs(df_ap_sim['Exponent'] - df_ap_sim['sim_exp'])
# %%
palette = 'deep'

f, ax = plt.subplots(figsize=(14, 14), ncols=2, nrows=2)
sns.lineplot(df_ap_sim, x='sim_freq', y='Exponent Δ', hue='method', palette=palette, legend=False, ax=ax[0, 0])
ax[0, 0].set_xlabel('Simulated Oscillation (Hz)')

sns.lineplot(df_ap_sim, x='sim_exp', y='Exponent Δ', hue='method', palette=palette, legend=False, ax=ax[0, 1])

ax[0, 1].set_xlabel('Simulated Exponent')

sns.lineplot(df_ap_sim, x='sim_bw', y='Exponent Δ', hue='method', palette=palette, legend=False, ax=ax[1, 0])

ax[1, 0].set_xlabel('Simulated Oscillation Width')

sns.lineplot(df_ap_sim, x='sim_height', y='Exponent Δ', hue='method', palette=palette, legend=False, ax=ax[1, 1])

ax[1, 1].set_xlabel('Simulated Oscillation Height')

sns.despine()
f.tight_layout()

# %%
df_pivot = df_ap_sim.reset_index().pivot_table(index=['method', 'sim_bw'], columns='sim_height', values='Exponent Δ')

# %%


def plot_mesh_bw_h(data, key, ax):
    cur_d = data.copy()
    data2plot = cur_d.query(f'method == "{key}"').to_numpy()
    x_labels = cur_d.query(f'method == "{key}"').columns
    y_labels = cur_d.query(f'method == "{key}"').droplevel('method').index
    title = key
    extent = [
        y_labels.min(),
        y_labels.max(),
        x_labels.min(),
        x_labels.max(),
    ]

    mesh = ax.imshow(
        data2plot, aspect=1.4, vmin=cur_d.min().min(), vmax=cur_d.max().max(), interpolation='none', extent=extent
    )
    ax.set_title(title)
    ax.set_xlabel('Oscillation \n Bandwidth')
    ax.set_ylabel('Oscillation \n Height')

    return mesh


# %%
f, axs = plt.subplots(ncols=2, figsize=(8, 18))
mesh = plot_mesh_bw_h(df_pivot, 'pyrasa', ax=axs[0])
mesh = plot_mesh_bw_h(df_pivot, 'specparam', ax=axs[1])

f.tight_layout()

cbar = f.colorbar(mesh, ax=axs.ravel().tolist(), orientation='vertical', aspect=4.0, fraction=0.05)
cbar.set_ticks([np.round(df_pivot.min().min(), 2), np.round(df_pivot.max().max(), 2)])
cbar.set_label('Exponent Δ')
f.savefig('../results/delta_exponent_sim_bw_h.svg')

# %%
df_nona = df_p_sim.copy().query('cf < 40').reset_index()
df_nona['Center \n Frequency Δ'] = np.abs(df_nona['cf'] - df_nona['sim_freq'])

# %%
f, ax = plt.subplots(figsize=(4, 6))
sns.violinplot(
    df_nona,
    x='method',
    y='Center \n Frequency Δ',
    hue='method',
    # split=True,
    legend=False,
    ax=ax,
)

ax.set_xlabel('')
sns.despine()

# %%
f, ax = plt.subplots(figsize=(14, 14), ncols=2, nrows=2)
sns.lineplot(df_nona, x='sim_freq', y='Center \n Frequency Δ', hue='method', palette=palette, legend=False, ax=ax[0, 0])
ax[0, 0].set_xlabel('Simulated Oscillation (Hz)')

sns.lineplot(df_nona, x='sim_exp', y='Center \n Frequency Δ', hue='method', palette=palette, legend=False, ax=ax[0, 1])

ax[0, 1].set_xlabel('Simulated Exponent')

sns.lineplot(df_nona, x='sim_bw', y='Center \n Frequency Δ', hue='method', palette=palette, legend=False, ax=ax[1, 0])

ax[1, 0].set_xlabel('Simulated Oscillation Width')

sns.lineplot(
    df_nona, x='sim_height', y='Center \n Frequency Δ', hue='method', palette=palette, legend=False, ax=ax[1, 1]
)

ax[1, 1].set_xlabel('Simulated Oscillation Height')

sns.despine()
f.tight_layout()


# %%

df_hp_sp = df_nona.query('method == "specparam"').query('pw > 1').copy()
df_ir = df_p_sim.query('method == "pyrasa"')

df_p_cmb = pd.concat([df_hp_sp, df_ir]).reset_index()

df_p_cmb['Center \n Frequency Δ'] = np.abs(df_p_cmb['cf'] - df_p_cmb['sim_freq'])

f, ax = plt.subplots(figsize=(4, 6))
sns.violinplot(
    df_p_cmb,
    x='method',
    y='Center \n Frequency Δ',
    hue='method',
    # split=True,
    legend=False,
    ax=ax,
)

ax.set_xlabel('')
sns.despine()

# %%
f, ax = plt.subplots(figsize=(14, 14), ncols=2, nrows=2)
sns.lineplot(
    df_p_cmb, x='sim_freq', y='Center \n Frequency Δ', hue='method', palette=palette, legend=False, ax=ax[0, 0]
)
ax[0, 0].set_xlabel('Simulated Oscillation (Hz)')

sns.lineplot(df_p_cmb, x='sim_exp', y='Center \n Frequency Δ', hue='method', palette=palette, legend=False, ax=ax[0, 1])

ax[0, 1].set_xlabel('Simulated Exponent')

sns.lineplot(df_p_cmb, x='sim_bw', y='Center \n Frequency Δ', hue='method', palette=palette, legend=False, ax=ax[1, 0])

ax[1, 0].set_xlabel('Simulated Oscillation Width')

sns.lineplot(
    df_p_cmb, x='sim_height', y='Center \n Frequency Δ', hue='method', palette=palette, legend=False, ax=ax[1, 1]
)

ax[1, 1].set_xlabel('Simulated Oscillation Height')

sns.despine()
f.tight_layout()

# %%
f, axes = plt.subplots(ncols=3, figsize=(14, 6))

sns.violinplot(
    df_ap_sim,
    x='method',
    y='Exponent Δ',
    hue='method',
    # split=True,
    legend=False,
    ax=axes[0],
)

sns.violinplot(
    df_nona,
    x='method',
    y='Center \n Frequency Δ',
    hue='method',
    # split=True,
    legend=False,
    ax=axes[1],
)

sns.violinplot(
    df_p_cmb,
    x='method',
    y='Center \n Frequency Δ',
    hue='method',
    # split=True,
    legend=False,
    ax=axes[2],
)

titles = ['', 'all peaks', 'specparam \n peaks > 1 (arbitrary)']
for ix, ax in enumerate(axes):
    ax.set_xlabel('')
    ax.set_title(titles[ix])

f.tight_layout()

sns.despine()
# %%



import numpy as np
import pytest
import scipy.signal as dsp

from pyrasa.utils.peak_utils import get_band_info, get_peak_params

from .settings import FS, MANY_OSC_FREQ


# Test slope fitting functionality
@pytest.mark.parametrize('osc_freq', MANY_OSC_FREQ, scope='session')
@pytest.mark.parametrize('fs', FS, scope='session')
def test_peak_detection(oscillation, fs, osc_freq):
    f_range = [1, 250]
    # test whether recombining periodic and aperiodic spectrum is equivalent to the original spectrum
    freqs, psd = dsp.welch(oscillation, fs, nperseg=int(4 * fs))
    freq_logical = np.logical_and(freqs >= f_range[0], freqs <= f_range[1])
    freqs, psd = freqs[freq_logical], psd[freq_logical]
    # test whether we can reconstruct the peaks correctly
    pe_params = get_peak_params(psd[np.newaxis, :], freqs, min_peak_height=0.1)
    assert bool(np.isclose(pe_params['cf'][0], osc_freq, atol=2))

    pe_filt = get_band_info(pe_params, freq_range=[osc_freq - 2, osc_freq + 2], ch_names=[])
    assert bool(np.isclose(pe_filt['cf'][0], osc_freq, atol=2))


@pytest.mark.parametrize('fs, exponent', [(500, -1)], scope='session')
def test_no_peak_detection(fixed_aperiodic_signal, fs):
    f_range = [1, 250]
    # test whether recombining periodic and aperiodic spectrum is equivalent to the original spectrum
    freqs, psd = dsp.welch(fixed_aperiodic_signal, fs, nperseg=int(4 * fs))
    freq_logical = np.logical_and(freqs >= f_range[0], freqs <= f_range[1])
    freqs, psd = freqs[freq_logical], psd[freq_logical]
    # test whether we can reconstruct the peaks correctly
    pe_params = get_peak_params(psd[np.newaxis, :], freqs, min_peak_height=0.1)
    assert pe_params.shape[0] == 0


@pytest.mark.parametrize('osc_freq, fs', [(10, 500)], scope='session')
def test_peak_detection_settings(oscillation, fs, osc_freq):
    f_range = [1, 250]
    # test whether recombining periodic and aperiodic spectrum is equivalent to the original spectrum
    freqs, psd = dsp.welch(oscillation, fs, nperseg=int(4 * fs))
    freq_logical = np.logical_and(freqs >= f_range[0], freqs <= f_range[1])
    freqs, psd = freqs[freq_logical], psd[freq_logical]
    psd_nan = psd.copy()
    psd_nan[10] = np.nan

    # test if nan in spectrum
    with pytest.raises(ValueError):
        get_peak_params(psd_nan[np.newaxis, :], freqs, min_peak_height=0.1)

    # test if cut spectrum runs
    get_peak_params(psd[np.newaxis, :], freqs, min_peak_height=0.1, cut_spectrum=[1, 40])

    # test if smooth False runs
    get_peak_params(psd[np.newaxis, :], freqs, min_peak_height=0.1, smooth=False)
